@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2006  The DOSBox Team
+ *  Copyright (C) 2002-2007  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: dos_mscdex.cpp,v 1.37 2006/03/26 11:54:44 qbix79 Exp $ */
+/* $Id: dos_mscdex.cpp,v 1.45 2007/01/21 16:21:22 c2woody Exp $ */
 
 #include <string.h>
 #include <ctype.h>
@@ -26,6 +26,7 @@
 #include "dos_inc.h"
 #include "setup.h"
 #include "support.h"
+#include "bios.h"
 
 #include "cdrom.h"
 
@@ -106,14 +107,17 @@ public:
 
 	bool		GetSubChannelData	(Bit8u subUnit, Bit8u& attr, Bit8u& track, Bit8u &index, TMSF& rel, TMSF& abs);
 
+	int			RemoveDrive			(Bit16u _drive);
 	int			AddDrive			(Bit16u _drive, char* physicalPath, Bit8u& subUnit);
+	bool 		HasDrive			(Bit16u drive);
+	void		ReplaceDrive		(CDROM_Interface* newCdrom, Bit8u subUnit);
 	void		GetDrives			(PhysPt data);
 	void		GetDriverInfo		(PhysPt data);
 	bool		GetVolumeName		(Bit8u subUnit, char* name);
 	bool		GetCopyrightName	(Bit16u drive, PhysPt data);
 	bool		GetAbstractName		(Bit16u drive, PhysPt data);
 	bool		GetDocumentationName(Bit16u drive, PhysPt data);
-	bool		GetDirectoryEntry	(Bit16u drive, bool copyFlag, PhysPt pathname, PhysPt buffer, Bitu& error);
+	bool		GetDirectoryEntry	(Bit16u drive, bool copyFlag, PhysPt pathname, PhysPt buffer, Bit16u& error);
 	bool		ReadVTOC			(Bit16u drive, Bit16u volume, PhysPt data, Bit16u& error);
 	bool		ReadSectors			(Bit16u drive, Bit32u sector, Bit16u num, PhysPt data);
 	bool		ReadSectors			(Bit8u subUnit, bool raw, Bit32u sector, Bit16u num, PhysPt data);
@@ -123,7 +127,7 @@ public:
 	bool		GetCDInfo			(Bit8u subUnit, Bit8u& tr1, Bit8u& tr2, TMSF& leadOut);
 	Bit32u		GetVolumeSize		(Bit8u subUnit);
 	bool		GetTrackInfo		(Bit8u subUnit, Bit8u track, Bit8u& attr, TMSF& start);
-	Bit16u		GetStatusWord		(Bit8u subUnit);
+	Bit16u		GetStatusWord		(Bit8u subUnit,Bit16u status);
 	bool		GetCurrentPos		(Bit8u subUnit, TMSF& pos);
 	Bit32u		GetDeviceStatus		(Bit8u subUnit);
 	bool		GetMediaStatus		(Bit8u subUnit, Bit8u& status);
@@ -196,10 +200,113 @@ Bit8u CMscdex::GetSubUnit(Bit16u _drive)
 	return 0xff;
 };
 
+int CMscdex::RemoveDrive(Bit16u _drive)
+{
+	Bit16u idx = MSCDEX_MAX_DRIVES;
+	for (Bit16u i=0; i<GetNumDrives(); i++) {
+		if (dinfo[i].drive == _drive) {
+			idx = i;
+			break;
+		}
+	}
+
+	if (idx == MSCDEX_MAX_DRIVES || (idx!=0 && idx!=GetNumDrives()-1)) return 0;
+	delete (cdrom)[idx];
+	if (idx==0) {
+		for (Bit16u i=0; i<GetNumDrives(); i++) {
+			if (i == MSCDEX_MAX_DRIVES-1) {
+				cdrom[i] = 0;
+				memset(&dinfo[i],0,sizeof(TDriveInfo));
+			} else {
+				dinfo[i] = dinfo[i+1];
+				cdrom[i] = cdrom[i+1];
+			}
+		}
+	} else {
+		cdrom[idx] = 0;
+		memset(&dinfo[idx],0,sizeof(TDriveInfo));
+	}
+	numDrives--;
+
+	if (GetNumDrives() == 0) {
+		DOS_DeviceHeader devHeader(PhysMake(rootDriverHeaderSeg,0));
+		Bit16u off = sizeof(DOS_DeviceHeader::sDeviceHeader);
+		devHeader.SetStrategy(off+4);		// point to the RETF (To deactivate MSCDEX)
+		devHeader.SetInterrupt(off+4);		// point to the RETF (To deactivate MSCDEX)
+		devHeader.SetDriveLetter(0);
+	} else if (idx==0) {
+		DOS_DeviceHeader devHeader(PhysMake(rootDriverHeaderSeg,0));
+		devHeader.SetDriveLetter(GetFirstDrive()+1);
+	}
+	return 1;
+}
+
 int CMscdex::AddDrive(Bit16u _drive, char* physicalPath, Bit8u& subUnit)
 {
 	subUnit = 0;
-	if (GetNumDrives()==0) {
+	if (GetNumDrives()+1>=MSCDEX_MAX_DRIVES) return 4;
+	if (GetNumDrives()) {
+		// Error check, driveletter have to be in a row
+		if (dinfo[0].drive-1!=_drive && dinfo[numDrives-1].drive+1!=_drive) 
+			return 1;
+	}
+	// Set return type to ok
+	int result = 0;
+	// Get Mounttype and init needed cdrom interface
+	switch (CDROM_GetMountType(physicalPath,forceCD)) {
+	case 0x00: {	
+		LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: Mounting physical cdrom: %s"	,physicalPath);
+#if defined (WIN32)
+		// Check OS
+		OSVERSIONINFO osi;
+		osi.dwOSVersionInfoSize = sizeof(osi);
+		GetVersionEx(&osi);
+		if ((osi.dwPlatformId==VER_PLATFORM_WIN32_NT) && (osi.dwMajorVersion>4)) {
+			// only WIN NT/200/XP
+			if (useCdromInterface==CDROM_USE_IOCTL) {
+				cdrom[numDrives] = new CDROM_Interface_Ioctl();
+				LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: IOCTL Interface.");
+				break;
+			}
+		}
+		if (useCdromInterface==CDROM_USE_ASPI) {
+		// all Wins - ASPI
+			cdrom[numDrives] = new CDROM_Interface_Aspi();
+			LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: ASPI Interface.");
+			break;
+		}
+#endif
+#if defined (LINUX) || defined(OS2)
+		// Always use IOCTL in Linux or OS/2
+		cdrom[numDrives] = new CDROM_Interface_Ioctl();
+		LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: IOCTL Interface.");
+#else
+		// Default case windows and other oses
+		cdrom[numDrives] = new CDROM_Interface_SDL();
+		LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: SDL Interface.");
+#endif
+		} break;
+	case 0x01:	// iso cdrom interface	
+		LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: Mounting iso file as cdrom: %s", physicalPath);
+		cdrom[numDrives] = new CDROM_Interface_Image((Bit8u)numDrives);
+		break;
+	case 0x02:	// fake cdrom interface (directories)
+		cdrom[numDrives] = new CDROM_Interface_Fake;
+		LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: Mounting directory as cdrom: %s",physicalPath);
+		LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: You wont have full MSCDEX support !");
+		result = 5;
+		break;
+	default	:	// weird result
+		return 6;
+	};
+
+	if (!cdrom[numDrives]->SetDevice(physicalPath,forceCD)) {
+//		delete cdrom[numDrives] ; mount seems to delete it
+		return 3;
+	}
+
+
+	if (rootDriverHeaderSeg==0) {
 		
 		Bit16u driverSize = sizeof(DOS_DeviceHeader::sDeviceHeader) + 10; // 10 = Bytes for 3 callbacks
 		
@@ -207,10 +314,22 @@ int CMscdex::AddDrive(Bit16u _drive, char* physicalPath, Bit8u& subUnit)
 		Bit16u seg = DOS_GetMemory(driverSize/16+((driverSize%16)>0));
 		DOS_DeviceHeader devHeader(PhysMake(seg,0));
 		devHeader.SetNextDeviceHeader	(0xFFFFFFFF);
+		devHeader.SetAttribute(0xc800);
 		devHeader.SetDriveLetter		(_drive+1);
 		devHeader.SetNumSubUnits		(1);
 		devHeader.SetName				("MSCD001 ");
 
+		//Link it in the device chain
+		Bit32u start = dos_infoblock.GetDeviceChain();
+		Bit16u segm  = start>>16;
+		Bit16u offm  = start&0xFFFF;
+		while(start != 0xFFFFFFFF) {
+			segm  = start>>16;
+			offm  = start&0xFFFF;
+			start = real_readd(segm,offm);
+		}
+		real_writed(segm,offm,seg<<16);
+
 		// Create Callback Strategy
 		Bit16u off = sizeof(DOS_DeviceHeader::sDeviceHeader);
 		Bitu call_strategy=CALLBACK_Allocate();
@@ -221,7 +340,8 @@ int CMscdex::AddDrive(Bit16u _drive, char* physicalPath, Bit8u& subUnit)
 		real_writeb(seg,off+4,(Bit8u)0xCB);		//A RETF Instruction
 		devHeader.SetStrategy(off);
 		
-		// Create Callback Interruptoff += 5;
+		// Create Callback Interrupt
+		off += 5;
 		Bitu call_interrupt=CALLBACK_Allocate();
 		CallBack_Handlers[call_interrupt]=MSCDEX_Interrupt_Handler;
 		real_writeb(seg,off+0,(Bit8u)0xFE);		//GRP 4
@@ -232,77 +352,48 @@ int CMscdex::AddDrive(Bit16u _drive, char* physicalPath, Bit8u& subUnit)
 		
 		rootDriverHeaderSeg = seg;
 	
-	} else {
-		// Error check, driveletter have to be in a row
-		if (dinfo[numDrives-1].drive+1!=_drive) return 1;
-	};
-
-	if (GetNumDrives()+1<MSCDEX_MAX_DRIVES) {
-		// Set return type to ok
-		int result = 0;
-		// Get Mounttype and init needed cdrom interface
-		switch (CDROM_GetMountType(physicalPath,forceCD)) {
-			case 0x00	: {	
-							LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: Mounting physical cdrom: %s"	,physicalPath);
-							#if defined (WIN32)
-								// Check OS
-								OSVERSIONINFO osi;
-								osi.dwOSVersionInfoSize = sizeof(osi);
-								GetVersionEx(&osi);
-								if ((osi.dwPlatformId==VER_PLATFORM_WIN32_NT) && (osi.dwMajorVersion>4)) {
-									// only WIN NT/200/XP
-									if (useCdromInterface==CDROM_USE_IOCTL) {
-										cdrom[numDrives] = new CDROM_Interface_Ioctl();
-										LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: IOCTL Interface.");
-										break;
-									}
-								}
-								if (useCdromInterface==CDROM_USE_ASPI) {
-									// all Wins - ASPI
-									cdrom[numDrives] = new CDROM_Interface_Aspi();
-									LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: ASPI Interface.");
-									break;
-								}
-							#endif
-							#if defined (LINUX) || defined(OS2)
-								// Always use IOCTL in Linux or OS/2
-//								if (useCdromInterface==CDROM_USE_IOCTL) {
-									cdrom[numDrives] = new CDROM_Interface_Ioctl();
-									LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: IOCTL Interface.");
-									break;
-//								}
-							#endif
-							cdrom[numDrives] = new CDROM_Interface_SDL();
-							LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: SDL Interface.");
-						  } break;
-			case 0x01	:	// iso cdrom interface	
-							LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: Mounting iso file as cdrom: %s", physicalPath);
-							cdrom[numDrives] = new CDROM_Interface_Image((Bit8u)numDrives);
-							break;
-			case 0x02	:	// fake cdrom interface (directories)
-							cdrom[numDrives] = new CDROM_Interface_Fake;
-							LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: Mounting directory as cdrom: %s",physicalPath);	
-							LOG(LOG_MISC,LOG_NORMAL)("MSCDEX: You wont have full MSCDEX support !");	
-							result = 5;
-							break;
-			default		:	// weird result
-							return 6;
-		};
-		if (!cdrom[numDrives]->SetDevice(physicalPath,forceCD)) return 3;
-		subUnit = (Bit8u)numDrives;
-		// Set drive
+	} else if (GetNumDrives() == 0) {
 		DOS_DeviceHeader devHeader(PhysMake(rootDriverHeaderSeg,0));
-		devHeader.SetNumSubUnits(devHeader.GetNumSubUnits()+1);
+		Bit16u off = sizeof(DOS_DeviceHeader::sDeviceHeader);
+		devHeader.SetDriveLetter(_drive+1);
+		devHeader.SetStrategy(off);
+		devHeader.SetInterrupt(off+5);
+	}
+
+	subUnit = (Bit8u)numDrives;
+	// Set drive
+	DOS_DeviceHeader devHeader(PhysMake(rootDriverHeaderSeg,0));
+	devHeader.SetNumSubUnits(devHeader.GetNumSubUnits()+1);
+
+	if (dinfo[0].drive-1==_drive) {
+		CDROM_Interface *_cdrom = cdrom[numDrives];
+		for (Bit16u i=GetNumDrives(); i>0; i--) {
+			dinfo[i] = dinfo[i-1];
+			cdrom[i] = cdrom[i-1];
+		}
+		cdrom[0] = _cdrom;
+		dinfo[0].drive		= (Bit8u)_drive;
+		dinfo[0].physDrive	= toupper(physicalPath[0]);
+	} else {
 		dinfo[numDrives].drive		= (Bit8u)_drive;
 		dinfo[numDrives].physDrive	= toupper(physicalPath[0]);
-		numDrives++;
-		// stop audio
-		StopAudio(subUnit);
-		return result;
 	}
-	return 4;
+	numDrives++;
+	// stop audio
+	StopAudio(subUnit);
+	return result;
 };
 
+bool CMscdex::HasDrive(Bit16u drive) {
+	return (GetSubUnit(drive) != 0xff);
+}
+
+void CMscdex::ReplaceDrive(CDROM_Interface* newCdrom, Bit8u subUnit) {
+	delete cdrom[subUnit];
+	cdrom[subUnit] = newCdrom;
+	StopAudio(subUnit);
+}
+
 PhysPt CMscdex::GetDefaultBuffer(void)
 {
 	if (defaultBufSeg==0) {
@@ -315,7 +406,7 @@ PhysPt CMscdex::GetDefaultBuffer(void)
 void CMscdex::GetDriverInfo	(PhysPt data)
 {
 	for (Bit16u i=0; i<GetNumDrives(); i++) {
-		mem_writeb(data  ,0x00);
+		mem_writeb(data  ,(Bit8u)i);	// subunit
 		mem_writed(data+1,RealMake(rootDriverHeaderSeg,0));
 		data+=5;
 	};
@@ -568,14 +659,14 @@ bool CMscdex::ReadSectors(Bit16u drive, Bit32u sector, Bit16u num, PhysPt data)
 	return ReadSectors(GetSubUnit(drive),false,sector,num,data);
 };
 
-bool CMscdex::GetDirectoryEntry(Bit16u drive, bool copyFlag, PhysPt pathname, PhysPt buffer, Bitu& error)
+bool CMscdex::GetDirectoryEntry(Bit16u drive, bool copyFlag, PhysPt pathname, PhysPt buffer, Bit16u& error)
 {
 	char	volumeID[6] = {0};
 	char	searchName[256];
 	char	entryName[256];
 	bool	foundComplete = false;
 	bool	foundName;
-	char*	useName;
+	char*	useName = 0;
 	Bitu	entryLength,nameLength;
 	// clear error
 	error = 0;
@@ -639,7 +730,6 @@ bool CMscdex::GetDirectoryEntry(Bit16u drive, bool copyFlag, PhysPt pathname, Ph
 		} while (index+33<=2048);
 		
 		if (foundName) {
-			// TO DO : name gefunden, Daten in den Buffer kopieren
 			if (foundComplete) {
 				if (copyFlag) {
 					LOG(LOG_MISC,LOG_WARN)("MSCDEX: GetDirEntry: Copyflag structure not entirely accurate maybe");
@@ -667,7 +757,7 @@ bool CMscdex::GetDirectoryEntry(Bit16u drive, bool copyFlag, PhysPt pathname, Ph
 				error = iso ? 1:0;
 				return true;
 			}
-			// directory wechseln
+			// change directory
 			dirEntrySector = mem_readd(defBuffer+index+2);
 			dirSize	= mem_readd(defBuffer+index+10);
 		} else {
@@ -715,10 +805,12 @@ Bit32u CMscdex::GetDeviceStatus(Bit8u subUnit)
 bool CMscdex::GetMediaStatus(Bit8u subUnit, Bit8u& status)
 {
 	if (subUnit>=numDrives) return false;
-	bool media,changed,open,result;
+/*	bool media,changed,open,result;
 	result = GetMediaStatus(subUnit,media,changed,open);
 	status = changed ? 0xFF : 0x01;
-	return result;
+	return result; */
+	status = getSwapRequest() ? 0xFF : 0x01;
+	return true;
 };
 
 bool CMscdex::LoadUnloadMedia(Bit8u subUnit, bool unload)
@@ -731,7 +823,7 @@ bool CMscdex::LoadUnloadMedia(Bit8u subUnit, bool unload)
 bool CMscdex::SendDriverRequest(Bit16u drive, PhysPt data)
 {
 	Bit8u subUnit = GetSubUnit(drive);
-	if (subUnit>=numDrives) return false;	
+	if (subUnit>=numDrives) return false;
 	// Get SubUnit
 	mem_writeb(data+1,subUnit);
 	// Call Strategy / Interrupt
@@ -740,13 +832,12 @@ bool CMscdex::SendDriverRequest(Bit16u drive, PhysPt data)
 	return true;
 };
 
-Bit16u CMscdex::GetStatusWord(Bit8u subUnit)
+Bit16u CMscdex::GetStatusWord(Bit8u subUnit,Bit16u status)
 {
-	if (subUnit>=numDrives) return 0x02; // error : Drive not ready
+	if (subUnit>=numDrives) return REQUEST_STATUS_ERROR | 0x02; // error : Drive not ready
 
-	Bit16u status ;
-	if (dinfo[subUnit].lastResult)	status = REQUEST_STATUS_DONE;				// ok
-	else							status = REQUEST_STATUS_ERROR; 
+	if (dinfo[subUnit].lastResult)	status |= REQUEST_STATUS_DONE;				// ok
+	else							status |= REQUEST_STATUS_ERROR; 
 
 	if (dinfo[subUnit].audioPlay) {
 		// Check if audio is still playing....
@@ -771,144 +862,169 @@ void CMscdex::InitNewMedia(Bit8u subUnit) {
 };
 
 static CMscdex* mscdex = 0;
+static PhysPt curReqheaderPtr = 0;
+
+static Bit16u MSCDEX_IOCTL_Input(PhysPt buffer,Bit8u drive_unit) {
+	Bitu ioctl_fct = mem_readb(buffer);
+	MSCDEX_LOG("MSCDEX: IOCTL INPUT Subfunction %02X",ioctl_fct);
+	switch (ioctl_fct) {
+		case 0x00 : /* Get Device Header address */
+					mem_writed(buffer+1,RealMake(mscdex->rootDriverHeaderSeg,0));
+					break;
+		case 0x01 :{/* Get current position */
+					TMSF pos;
+					mscdex->GetCurrentPos(drive_unit,pos);
+					Bit8u addr_mode = mem_readb(buffer+1);
+					if (addr_mode==0) {			// HSG
+						mem_writed(buffer+2,MSF_TO_FRAMES (pos.min, pos.sec, pos.fr));
+					} else if (addr_mode==1) {	// Red book
+						mem_writeb(buffer+2,pos.fr);
+						mem_writeb(buffer+3,pos.sec);
+						mem_writeb(buffer+4,pos.min);
+						mem_writeb(buffer+5,0x00);
+					} else {
+						LOG_MSG("MSCDEX: Get position: invalid address mode %x",addr_mode);
+						return 0x03;		// invalid function
+					}
+				   }break;
+		case 0x06 : /* Get Device status */
+					mem_writed(buffer+1,mscdex->GetDeviceStatus(drive_unit)); 
+					break;
+		case 0x07 : /* Get sector size */
+					if (mem_readb(buffer+1)==0) mem_writed(buffer+2,2048);
+					else if (mem_readb(buffer+1)==1) mem_writed(buffer+2,2352);
+					else return 0x03;		// invalid function
+					break;
+		case 0x08 : /* Get size of current volume */
+					mem_writed(buffer+1,mscdex->GetVolumeSize(drive_unit));
+					break;
+		case 0x09 : /* Media change ? */
+					Bit8u status;
+					if (!mscdex->GetMediaStatus(drive_unit,status)) {
+						status = 0;		// state unknown
+					}
+					mem_writeb(buffer+1,status);
+					break;
+		case 0x0A : /* Get Audio Disk info */	
+					Bit8u tr1,tr2; TMSF leadOut;
+					mscdex->GetCDInfo(drive_unit,tr1,tr2,leadOut);
+					mem_writeb(buffer+1,tr1);
+					mem_writeb(buffer+2,tr2);
+					mem_writeb(buffer+3,leadOut.fr);
+					mem_writeb(buffer+4,leadOut.sec);
+					mem_writeb(buffer+5,leadOut.min);
+					mem_writeb(buffer+6,0x00);
+					break;
+		case 0x0B :{/* Audio Track Info */
+					Bit8u attr; TMSF start;
+					Bit8u track = mem_readb(buffer+1);
+					mscdex->GetTrackInfo(drive_unit,track,attr,start);		
+					mem_writeb(buffer+2,start.fr);
+					mem_writeb(buffer+3,start.sec);
+					mem_writeb(buffer+4,start.min);
+					mem_writeb(buffer+5,0x00);
+					mem_writeb(buffer+6,attr);
+					break; };
+		case 0x0C :{/* Get Audio Sub Channel data */
+					Bit8u attr,track,index; 
+					TMSF abs,rel;
+					mscdex->GetSubChannelData(drive_unit,attr,track,index,rel,abs);
+					mem_writeb(buffer+1,attr);
+					mem_writeb(buffer+2,track);
+					mem_writeb(buffer+3,index);
+					mem_writeb(buffer+4,rel.min);
+					mem_writeb(buffer+5,rel.sec);
+					mem_writeb(buffer+6,rel.fr);
+					mem_writeb(buffer+7,0x00);
+					mem_writeb(buffer+8,abs.min);
+					mem_writeb(buffer+9,abs.sec);
+					mem_writeb(buffer+10,abs.fr);
+					break;
+				   };
+		case 0x0E :{ /* Get UPC */	
+					Bit8u attr; char upc[8];
+					mscdex->GetUPC(drive_unit,attr,&upc[0]);
+					mem_writeb(buffer+1,attr);
+					for (int i=0; i<7; i++) mem_writeb(buffer+2+i,upc[i]);
+					mem_writeb(buffer+9,0x00);
+					break;
+				   };
+		case 0x0F :{ /* Get Audio Status */	
+					bool playing,pause;
+					TMSF resStart,resEnd;
+					mscdex->GetAudioStatus(drive_unit,playing,pause,resStart,resEnd);
+					mem_writeb(buffer+1,pause);
+					mem_writeb(buffer+3,resStart.min);
+					mem_writeb(buffer+4,resStart.sec);
+					mem_writeb(buffer+5,resStart.fr);
+					mem_writeb(buffer+6,0x00);
+					mem_writeb(buffer+7,resEnd.min);
+					mem_writeb(buffer+8,resEnd.sec);
+					mem_writeb(buffer+9,resEnd.fr);
+					mem_writeb(buffer+10,0x00);
+					break;
+				   };
+		default :	LOG(LOG_MISC,LOG_ERROR)("MSCDEX: Unsupported IOCTL INPUT Subfunction %02X",ioctl_fct);
+					return 0x03;	// invalid function
+	}
+	return 0x00;	// success
+}
+
+static Bit16u MSCDEX_IOCTL_Optput(PhysPt buffer,Bit8u drive_unit) {
+	Bitu ioctl_fct = mem_readb(buffer);
+//	MSCDEX_LOG("MSCDEX: IOCTL OUTPUT Subfunction %02X",ioctl_fct);
+	switch (ioctl_fct) {
+		case 0x00 :	// Unload /eject media
+					if (!mscdex->LoadUnloadMedia(drive_unit,true)) return 0x02;
+					break;
+		case 0x03: //Audio Channel control
+					MSCDEX_LOG("MSCDEX: Audio Channel Control used. Not handled. Faking succes!");
+		case 0x01 : // (un)Lock door 
+					// do nothing -> report as success
+					break;
+		case 0x02 : // Reset Drive
+					LOG(LOG_MISC,LOG_WARN)("cdromDrive reset");
+					if (!mscdex->StopAudio(drive_unit))  return 0x02;
+					break;
+		case 0x05 :	// load media
+					if (!mscdex->LoadUnloadMedia(drive_unit,false)) return 0x02;
+		default	:	LOG(LOG_MISC,LOG_ERROR)("MSCDEX: Unsupported IOCTL OUTPUT Subfunction %02X",ioctl_fct);
+					return 0x03;	// invalid function
+	}
+	return 0x00;	// success
+}
 
 static Bitu MSCDEX_Strategy_Handler(void) {
-//	MSCDEX_LOG("MSCDEX: Device Strategy Routine called.");
+	curReqheaderPtr = PhysMake(SegValue(es),reg_bx);
+//	MSCDEX_LOG("MSCDEX: Device Strategy Routine called, request header at %x",curReqheaderPtr);
 	return CBRET_NONE;
 }
 
 static Bitu MSCDEX_Interrupt_Handler(void) {
-	Bit8u	subFuncNr	= 0xFF;
-	PhysPt	data		= PhysMake(SegValue(es),reg_bx);
-	Bit8u	subUnit		= mem_readb(data+1);
-	Bit8u	funcNr		= mem_readb(data+2);
+	if (curReqheaderPtr==0) {
+		MSCDEX_LOG("MSCDEX: invalid call to interrupt handler");						
+		return CBRET_NONE;
+	}
+	Bit8u	subUnit		= mem_readb(curReqheaderPtr+1);
+	Bit8u	funcNr		= mem_readb(curReqheaderPtr+2);
+	Bit16u	errcode		= 0;
+	PhysPt	buffer		= 0;
 
 	MSCDEX_LOG("MSCDEX: Driver Function %02X",funcNr);
 
+	if ((funcNr==0x03) || (funcNr==0x0c) || (funcNr==0x80) || (funcNr==0x82)) {
+		buffer = PhysMake(mem_readw(curReqheaderPtr+0x10),mem_readw(curReqheaderPtr+0x0E));
+	}
+
  	switch (funcNr) {
-	
 		case 0x03	: {	/* IOCTL INPUT */
-						PhysPt buffer	= PhysMake(mem_readw(data+0x10),mem_readw(data+0x0E));
-						subFuncNr		= mem_readb(buffer);
-						MSCDEX_LOG("MSCDEX: IOCTL INPUT Subfunction %02X",subFuncNr);
-						switch (subFuncNr) {
-							case 0x00 : /* Get Device Header address */
-										mem_writed(buffer+1,RealMake(mscdex->rootDriverHeaderSeg,0));
-										break;
-							case 0x01 :{/* Get current position */
-										TMSF pos;
-										mscdex->GetCurrentPos(subUnit,pos);
-										/*mem_writeb(buffer+1,0x01); // Red book
-										mem_writeb(buffer+2,pos.fr);
-										mem_writeb(buffer+3,pos.sec);
-										mem_writeb(buffer+4,pos.min);
-										mem_writeb(buffer+5,0x00);*/
-							//Changed to HSG as default 
-							//(Seems to fix a few broken games which don't test for it)
-										mem_writeb(buffer+1,0x00); //HSG
-										mem_writed(buffer+2,MSF_TO_FRAMES (pos.min, pos.sec, pos.fr));
-									   }break;
-							case 0x06 : /* Get Device status */
-										mem_writed(buffer+1,mscdex->GetDeviceStatus(subUnit)); 
-										break;
-							case 0x07 : /* Get sector size */
-										if (mem_readb(buffer+1)==0x01)	mem_writed(buffer+2,2352);
-										else							mem_writed(buffer+2,2048);
-										break;
-							case 0x08 : /* Get size of current volume */
-										mem_writed(buffer+1,mscdex->GetVolumeSize(subUnit));
-										break;
-							case 0x09 : /* Media change ? */
-										Bit8u status;
-										//TEMP mscdex->GetMediaStatus(subUnit,status);
-										status = 1;
-										mem_writeb(buffer+1,status);
-										break;
-							case 0x0A : /* Get Audio Disk info */	
-										Bit8u tr1,tr2; TMSF leadOut;
-										mscdex->GetCDInfo(subUnit,tr1,tr2,leadOut);
-										mem_writeb(buffer+1,tr1);
-										mem_writeb(buffer+2,tr2);
-										mem_writeb(buffer+3,leadOut.fr);
-										mem_writeb(buffer+4,leadOut.sec);
-										mem_writeb(buffer+5,leadOut.min);
-										mem_writeb(buffer+6,0x00);
-										break;
-							case 0x0B :{/* Audio Track Info */
-										Bit8u attr; TMSF start;
-										Bit8u track = mem_readb(buffer+1);
-										mscdex->GetTrackInfo(subUnit,track,attr,start);		
-										mem_writeb(buffer+2,start.fr);
-										mem_writeb(buffer+3,start.sec);
-										mem_writeb(buffer+4,start.min);
-										mem_writeb(buffer+5,0x00);
-										mem_writeb(buffer+6,attr);
-										break; };
-							case 0x0C :{/* Get Audio Sub Channel data */
-										Bit8u attr,track,index; 
-										TMSF abs,rel;
-										mscdex->GetSubChannelData(subUnit,attr,track,index,rel,abs);
-										mem_writeb(buffer+1,attr);
-										mem_writeb(buffer+2,track);
-										mem_writeb(buffer+3,index);
-										mem_writeb(buffer+4,rel.min);
-										mem_writeb(buffer+5,rel.sec);
-										mem_writeb(buffer+6,rel.fr);
-										mem_writeb(buffer+7,0x00);
-										mem_writeb(buffer+8,abs.min);
-										mem_writeb(buffer+9,abs.sec);
-										mem_writeb(buffer+10,abs.fr);
-										break;
-									   };
-							case 0x0E :{ /* Get UPC */	
-										Bit8u attr; char upc[8];
-										mscdex->GetUPC(subUnit,attr,&upc[0]);
-										mem_writeb(buffer+1,attr);
-										for (int i=0; i<7; i++) mem_writeb(buffer+2+i,upc[i]);
-										mem_writeb(buffer+9,0x00);
-										break;
-									   };
-							case 0x0F :{ /* Get Audio Status */	
-										bool playing,pause;
-										TMSF resStart,resEnd;
-										mscdex->GetAudioStatus(subUnit,playing,pause,resStart,resEnd);
-										mem_writeb(buffer+1,pause);
-										mem_writeb(buffer+3,resStart.min);
-										mem_writeb(buffer+4,resStart.sec);
-										mem_writeb(buffer+5,resStart.fr);
-										mem_writeb(buffer+6,0x00);
-										mem_writeb(buffer+7,resEnd.min);
-										mem_writeb(buffer+8,resEnd.sec);
-										mem_writeb(buffer+9,resEnd.fr);
-										mem_writeb(buffer+10,0x00);
-										break;
-									   };
-							default :	LOG(LOG_MISC,LOG_ERROR)("MSCDEX: Unsupported IOCTL INPUT Subfunction %02X",subFuncNr);
-										break;
-						}
+						Bit16u error=MSCDEX_IOCTL_Input(buffer,subUnit);
+						if (error) errcode = error;
 						break;
 					  };
 		case 0x0C	: {	/* IOCTL OUTPUT */
-						PhysPt buffer	= PhysMake(mem_readw(data+0x10),mem_readw(data+0x0E));
-						subFuncNr		= mem_readb(buffer);
-//						LOG(LOG_MISC,LOG_ERROR)("MSCDEX: IOCTL OUTPUT Subfunction %02X",subFuncNr);
-						switch (subFuncNr) {
-							case 0x00 :	// Unload /eject) media
-										mscdex->LoadUnloadMedia(subUnit,true);
-										break;
-							case 0x01 : // (un)Lock door 
-										// do nothing -> report as success
-										break;
-							case 0x02 : // Reset Drive
-										LOG(LOG_MISC,LOG_WARN)("cdromDrive reset");
-										mscdex->StopAudio(subUnit);
-										break;
-							case 0x05 :	// load media
-										mscdex->LoadUnloadMedia(subUnit,false);
-										break;
-							default	:	LOG(LOG_MISC,LOG_ERROR)("MSCDEX: Unsupported IOCTL OUTPUT Subfunction %02X",subFuncNr);
-										break;
-						};
+						Bit16u error=MSCDEX_IOCTL_Optput(buffer,subUnit);
+						if (error) errcode = error;
 						break;
 					  };
 		case 0x0D	:	// device open
@@ -916,22 +1032,21 @@ static Bitu MSCDEX_Interrupt_Handler(void) {
 						break;
 		case 0x80	:	// Read long
 		case 0x82	: { // Read long prefetch -> both the same here :)
-						PhysPt buff  = PhysMake(mem_readw(data+0x10),mem_readw(data+0x0E));
-						Bit32u start = mem_readd(data+0x14);
-						Bit16u len	 = mem_readw(data+0x12);
-						bool raw	 = (mem_readb(data+0x18)==1);
-						if (mem_readb(data+0x0D)==0x00) // HSG
-							mscdex->ReadSectors(subUnit,raw,start,len,buff);
+						Bit32u start = mem_readd(curReqheaderPtr+0x14);
+						Bit16u len	 = mem_readw(curReqheaderPtr+0x12);
+						bool raw	 = (mem_readb(curReqheaderPtr+0x18)==1);
+						if (mem_readb(curReqheaderPtr+0x0D)==0x00) // HSG
+							mscdex->ReadSectors(subUnit,raw,start,len,buffer);
 						else 
-							mscdex->ReadSectorsMSF(subUnit,raw,start,len,buff);
+							mscdex->ReadSectorsMSF(subUnit,raw,start,len,buffer);
 						break;
 					  };
 		case 0x83	:	// Seek - dont care :)
 						break;
 		case 0x84	: {	/* Play Audio Sectors */
-						Bit32u start = mem_readd(data+0x0E);
-						Bit32u len	 = mem_readd(data+0x12);
-						if (mem_readb(data+0x0D)==0x00) // HSG
+						Bit32u start = mem_readd(curReqheaderPtr+0x0E);
+						Bit32u len	 = mem_readd(curReqheaderPtr+0x12);
+						if (mem_readb(curReqheaderPtr+0x0D)==0x00) // HSG
 							mscdex->PlayAudioSector(subUnit,start,len);
 						else // RED BOOK
 							mscdex->PlayAudioMSF(subUnit,start,len);
@@ -949,8 +1064,8 @@ static Bitu MSCDEX_Interrupt_Handler(void) {
 	};
 	
 	// Set Statusword
-	mem_writew(data+3,mscdex->GetStatusWord(subUnit));
-	MSCDEX_LOG("MSCDEX: Status : %04X",mem_readw(data+3));						
+	mem_writew(curReqheaderPtr+3,mscdex->GetStatusWord(subUnit,errcode));
+	MSCDEX_LOG("MSCDEX: Status : %04X",mem_readw(curReqheaderPtr+3));						
 	return CBRET_NONE;
 }
 
@@ -967,7 +1082,7 @@ static bool MSCDEX_Handler(void) {
 		}
 	}
 
-	if (reg_ah!=0x15) return false;
+	if (reg_ah!=0x15) return false;		// not handled here, continue chain
 
 	PhysPt data = PhysMake(SegValue(es),reg_bx);
 	MSCDEX_LOG("MSCDEX: INT 2F %04X BX= %04X CX=%04X",reg_ax,reg_bx,reg_cx);
@@ -1039,43 +1154,79 @@ static bool MSCDEX_Handler(void) {
 						mscdex->GetDrives(data);
 						return true;
 		case 0x150F: {	// Get directory entry
-						Bitu error;
+						Bit16u error;
 						bool success = mscdex->GetDirectoryEntry(reg_cl,reg_ch&1,data,PhysMake(reg_si,reg_di),error);
 						reg_ax = error;
 						CALLBACK_SCF(!success);
 					 };	return true;
 		case 0x1510:	/* Device driver request */
-						mscdex->SendDriverRequest(reg_cx,data);
+						if (mscdex->SendDriverRequest(reg_cx,data)) {
+							CALLBACK_SCF(false);
+						} else {
+							reg_ax = 0x0f;	// invalid drive
+							CALLBACK_SCF(true);
+						}
 						return true;
- 
-		default	:		LOG(LOG_MISC,LOG_ERROR)("MSCDEX: Unknwon call : %04X",reg_ax);
-						return true;
-
 	};
-	return false;
+	LOG(LOG_MISC,LOG_ERROR)("MSCDEX: Unknwon call : %04X",reg_ax);
+	return true;
 };
 
 class device_MSCDEX : public DOS_Device {
 public:
 	device_MSCDEX() { SetName("MSCD001"); }
-	bool Read (Bit8u * data,Bit16u * size) { return false;}
-	bool Write(Bit8u * data,Bit16u * size) { 
+	bool Read (Bit8u * /*data*/,Bit16u * /*size*/) { return false;}
+	bool Write(Bit8u * /*data*/,Bit16u * /*size*/) { 
 		LOG(LOG_ALL,LOG_NORMAL)("Write to mscdex device");	
 		return false;
 	}
-	bool Seek(Bit32u * pos,Bit32u type){return false;}
+	bool Seek(Bit32u * /*pos*/,Bit32u /*type*/){return false;}
 	bool Close(){return false;}
-	Bit16u GetInformation(void){return 0x8093;}
+	Bit16u GetInformation(void){return 0xc880;}
+	bool ReadFromControlChannel(PhysPt bufptr,Bit16u size,Bit16u * retcode);
+	bool WriteToControlChannel(PhysPt bufptr,Bit16u size,Bit16u * retcode);
 private:
-	Bit8u cache;
+//	Bit8u cache;
 };
 
+bool device_MSCDEX::ReadFromControlChannel(PhysPt bufptr,Bit16u size,Bit16u * retcode) { 
+	if (MSCDEX_IOCTL_Input(bufptr,0)==0) {
+		*retcode=size;
+		return true;
+	}
+	return false;
+}
+
+bool device_MSCDEX::WriteToControlChannel(PhysPt bufptr,Bit16u size,Bit16u * retcode) { 
+	if (MSCDEX_IOCTL_Optput(bufptr,0)==0) {
+		*retcode=size;
+		return true;
+	}
+	return false;
+}
+
 int MSCDEX_AddDrive(char driveLetter, const char* physicalPath, Bit8u& subUnit)
 {
 	int result = mscdex->AddDrive(driveLetter-'A',(char*)physicalPath,subUnit);
 	return result;
 };
 
+int MSCDEX_RemoveDrive(char driveLetter)
+{
+	if(!mscdex) return 0;
+	return mscdex->RemoveDrive(driveLetter-'A');
+};
+
+bool MSCDEX_HasDrive(char driveLetter)
+{
+	return mscdex->HasDrive(driveLetter-'A');
+};
+
+void MSCDEX_ReplaceDrive(CDROM_Interface* cdrom, Bit8u subUnit)
+{
+	mscdex->ReplaceDrive(cdrom, subUnit);
+};
+
 bool MSCDEX_GetVolumeName(Bit8u subUnit, char* name)
 {
 	return mscdex->GetVolumeName(subUnit,name);
@@ -1113,7 +1264,9 @@ void MSCDEX_SetCDInterface(int intNr, int numCD)
 
 void MSCDEX_ShutDown(Section* sec)
 {
-	delete mscdex; mscdex = 0;
+	delete mscdex;
+	mscdex = 0;
+	curReqheaderPtr = 0;
 };
 
 void MSCDEX_Init(Section* sec) 
@@ -1123,6 +1276,7 @@ void MSCDEX_Init(Section* sec)
 	/* Register the mscdex device */
 	DOS_Device * newdev = new device_MSCDEX();
 	DOS_AddDevice(newdev);
+	curReqheaderPtr = 0;
 	/* Add Multiplexer */
 	DOS_AddMultiplexHandler(MSCDEX_Handler);
 	/* Create MSCDEX */
