@@ -1,6 +1,6 @@
 
 /*
- *  Copyright (C) 2002-2006  The DOSBox Team
+ *  Copyright (C) 2002-2007  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: drive_cache.cpp,v 1.46 2006/02/09 11:47:48 qbix79 Exp $ */
+/* $Id: drive_cache.cpp,v 1.49 2007/01/08 19:45:39 qbix79 Exp $ */
 
 #include "drives.h"
 #include "dos_inc.h"
@@ -690,11 +690,13 @@ bool DOS_Drive_Cache::SetResult(CFileInfo* dir, char* &result, Bit16u entryNr)
 bool DOS_Drive_Cache::FindFirst(char* path, Bitu& id)
 {
 	Bit16u	dirID;
-	Bitu	dirFindFirstID = this->nextFreeFindFirst++; //increase it for the next search
+	Bitu	dirFindFirstID = this->nextFreeFindFirst;
 
 	// Cache directory in 
 	if (!OpenDir(path,dirID)) return false;
 
+	this->nextFreeFindFirst++; //increase it for the next search
+
 	if (dirFindFirstID == MAX_OPENDIRS) {
 		// no free slot found...
 		LOG(LOG_MISC,LOG_ERROR)("DIRCACHE: FindFirst/Next: All slots full. Resetting");
@@ -745,43 +747,3 @@ bool DOS_Drive_Cache::FindNext(Bitu id, char* &result)
 	}
 	return true;
 };
-
-// ****************************************************************************
-// No Dir Cache, 
-// ****************************************************************************
-
-static DIR*	srch_opendir = 0;
-
-DOS_No_Drive_Cache::DOS_No_Drive_Cache(const char* path)
-{
-	SetBaseDir(path);
-};
-
-void DOS_No_Drive_Cache::SetBaseDir(const char* path)
-{
-	strcpy(basePath,path);
-}
-
-bool DOS_No_Drive_Cache::OpenDir(const char* path, Bit16u& id)
-{
-	id = 0;
-	strcpy(dirPath,path);
-	if((srch_opendir=opendir(dirPath))==NULL) return false;
-	return true;
-};
-
-bool DOS_No_Drive_Cache::ReadDir(Bit16u id, char* &result)
-{
-	static char res[CROSS_LEN] = { 0 };
-	dirent * ent;		
-
-	if (!srch_opendir) return false;
-	if ((ent=readdir(srch_opendir))==NULL) {
-		strcpy(res,ent->d_name);
-		result=res;
-		closedir(srch_opendir);
-		srch_opendir=NULL;
-		return false;
-	}
-	return true;
-};
