@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2004  The DOSBox Team
+ *  Copyright (C) 2002-2007  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: drive_fat.cpp,v 1.6 2004/11/03 23:13:54 qbix79 Exp $ */
+/* $Id: drive_fat.cpp,v 1.20 2007/01/08 19:45:39 qbix79 Exp $ */
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -37,6 +37,9 @@
 #define FAT16		   1
 #define FAT32		   2
 
+Bit8u fatSectBuffer[1024];
+Bit32u curFatSect;
+
 class fatFile : public DOS_File {
 public:
 	fatFile(const char* name, Bit32u startCluster, Bit32u fileLen, fatDrive *useDrive);
@@ -69,9 +72,8 @@ private:
 /* OUT - char * filearray: Name in DOS directory format, eleven char, e.g. bob     txt */
 static void convToDirFile(char *filename, char *filearray) {
 	Bit32u charidx = 0;
-	Bit32u flen;
-	int i;
-	flen = strlen(filename);
+	Bit32u flen,i;
+	flen = (Bit32u)strlen(filename);
 	memset(filearray, 32, 11);
 	for(i=0;i<flen;i++) {
 		if(charidx >= 11) break;
@@ -82,24 +84,22 @@ static void convToDirFile(char *filename, char *filearray) {
 			charidx = 8;
 		}
 	}
-
-	
-
 }
 
-fatFile::fatFile(const char* name, Bit32u startCluster, Bit32u fileLen, fatDrive *useDrive) {
+fatFile::fatFile(const char* /*name*/, Bit32u startCluster, Bit32u fileLen, fatDrive *useDrive) {
 	Bit32u seekto = 0;
 	firstCluster = startCluster;
 	myDrive = useDrive;
 	filelength = fileLen;
+	open = true;
 	loadedSector = false;
 	curSectOff = 0;
 	seekpos = 0;
 	memset(&sectorBuffer[0], 0, sizeof(sectorBuffer));
 	
 	if(filelength > 0) {
-	Seek(&seekto, DOS_SEEK_SET);
-	myDrive->loadedDisk->Read_AbsoluteSector(currentSector, sectorBuffer);
+		Seek(&seekto, DOS_SEEK_SET);
+		myDrive->loadedDisk->Read_AbsoluteSector(currentSector, sectorBuffer);
 		loadedSector = true;
 	}
 }
@@ -111,6 +111,19 @@ bool fatFile::Read(Bit8u * data, Bit16u *size) {
 		return true;
 	}
 
+	if (!loadedSector) {
+		currentSector = myDrive->getAbsoluteSectFromBytePos(firstCluster, seekpos);
+		if(currentSector == 0) {
+			/* EOC reached before EOF */
+			*size = 0;
+			loadedSector = false;
+			return true;
+		}
+		curSectOff = 0;
+		myDrive->loadedDisk->Read_AbsoluteSector(currentSector, sectorBuffer);
+		loadedSector = true;
+	}
+
 	sizedec = *size;
 	sizecount = 0;
 	while(sizedec != 0) {
@@ -118,23 +131,23 @@ bool fatFile::Read(Bit8u * data, Bit16u *size) {
 			*size = sizecount;
 			return true; 
 		}
-		data[sizecount] = sectorBuffer[curSectOff];
-		curSectOff++;
+		data[sizecount++] = sectorBuffer[curSectOff++];
 		seekpos++;
 		if(curSectOff >= myDrive->getSectorSize()) {
 			currentSector = myDrive->getAbsoluteSectFromBytePos(firstCluster, seekpos);
 			if(currentSector == 0) {
 				/* EOC reached before EOF */
-				LOG_MSG("EOC reached before EOF, seekpos %d, filelen %d", seekpos, filelength);
+				//LOG_MSG("EOC reached before EOF, seekpos %d, filelen %d", seekpos, filelength);
 				*size = sizecount;
+				loadedSector = false;
 				return true;
 			}
 			curSectOff = 0;
 			myDrive->loadedDisk->Read_AbsoluteSector(currentSector, sectorBuffer);
-			LOG_MSG("Reading absolute sector at %d for seekpos %d", currentSector, seekpos);
+			loadedSector = true;
+			//LOG_MSG("Reading absolute sector at %d for seekpos %d", currentSector, seekpos);
 		}
 		--sizedec;
-		sizecount++;
 	}
 	*size =sizecount;
 	return true;
@@ -147,17 +160,37 @@ bool fatFile::Write(Bit8u * data, Bit16u *size) {
 	Bit16u sizedec, sizecount;
 	sizedec = *size;
 	sizecount = 0;
+
 	while(sizedec != 0) {
 		/* Increase filesize if necessary */
 		if(seekpos >= filelength) {
 			if(filelength == 0) {
 				firstCluster = myDrive->getFirstFreeClust();
 				myDrive->allocateCluster(firstCluster, 0);
+				currentSector = myDrive->getAbsoluteSectFromBytePos(firstCluster, seekpos);
+				myDrive->loadedDisk->Read_AbsoluteSector(currentSector, sectorBuffer);
+				loadedSector = true;
 			}
 			filelength = seekpos+1;
+			if (!loadedSector) {
+				currentSector = myDrive->getAbsoluteSectFromBytePos(firstCluster, seekpos);
+				if(currentSector == 0) {
+					/* EOC reached before EOF - try to increase file allocation */
+					myDrive->appendCluster(firstCluster);
+					/* Try getting sector again */
+					currentSector = myDrive->getAbsoluteSectFromBytePos(firstCluster, seekpos);
+					if(currentSector == 0) {
+						/* No can do. lets give up and go home.  We must be out of room */
+						goto finalizeWrite;
+					}
+				}
+				curSectOff = 0;
+				myDrive->loadedDisk->Read_AbsoluteSector(currentSector, sectorBuffer);
+
+				loadedSector = true;
+			}
 		}
-		sectorBuffer[curSectOff] = data[sizecount];
-		curSectOff++;
+		sectorBuffer[curSectOff++] = data[sizecount++];
 		seekpos++;
 		if(curSectOff >= myDrive->getSectorSize()) {
 			if(loadedSector) myDrive->loadedDisk->Write_AbsoluteSector(currentSector, sectorBuffer);
@@ -170,8 +203,8 @@ bool fatFile::Write(Bit8u * data, Bit16u *size) {
 				currentSector = myDrive->getAbsoluteSectFromBytePos(firstCluster, seekpos);
 				if(currentSector == 0) {
 					/* No can do. lets give up and go home.  We must be out of room */
+					loadedSector = false;
 					goto finalizeWrite;
-				
 				}
 			}
 			curSectOff = 0;
@@ -180,22 +213,21 @@ bool fatFile::Write(Bit8u * data, Bit16u *size) {
 			loadedSector = true;
 		}
 		--sizedec;
-		sizecount++;
 	}
+	if(curSectOff>0 && loadedSector) myDrive->loadedDisk->Write_AbsoluteSector(currentSector, sectorBuffer);
 
 finalizeWrite:
 	myDrive->directoryBrowse(dirCluster, &tmpentry, dirIndex);
 	tmpentry.entrysize = filelength;
-	tmpentry.loFirstClust = firstCluster;
+	tmpentry.loFirstClust = (Bit16u)firstCluster;
 	myDrive->directoryChange(dirCluster, &tmpentry, dirIndex);
 
 	*size =sizecount;
 	return true;
-
 }
 
 bool fatFile::Seek(Bit32u *pos, Bit32u type) {
-	Bit32s seekto;
+	Bit32s seekto=0;
 	
 	switch(type) {
 		case DOS_SEEK_SET:
@@ -209,14 +241,19 @@ bool fatFile::Seek(Bit32u *pos, Bit32u type) {
 			seekto = (Bit32s)filelength + (Bit32s)*pos;
 			break;
 	}
-	LOG_MSG("Seek to %d with type %d (absolute value %d)", *pos, type, seekto);
+//	LOG_MSG("Seek to %d with type %d (absolute value %d)", *pos, type, seekto);
 
 	if((Bit32u)seekto > filelength) seekto = (Bit32s)filelength;
 	if(seekto<0) seekto = 0;
 	seekpos = (Bit32u)seekto;
 	currentSector = myDrive->getAbsoluteSectFromBytePos(firstCluster, seekpos);
-	curSectOff = seekpos % myDrive->getSectorSize();
-	myDrive->loadedDisk->Read_AbsoluteSector(currentSector, sectorBuffer);
+	if (currentSector == 0) {
+		/* not within file size, thus no sector is available */
+		loadedSector = false;
+	} else {
+		curSectOff = seekpos % myDrive->getSectorSize();
+		myDrive->loadedDisk->Read_AbsoluteSector(currentSector, sectorBuffer);
+	}
 	*pos = seekpos;
 	return true;
 }
@@ -229,7 +266,7 @@ bool fatFile::Close() {
 }
 
 Bit16u fatFile::GetInformation(void) {
-	return 0x202;
+	return 0;
 }
 
 bool fatFile::UpdateDateTimeFromHost(void) {
@@ -241,15 +278,10 @@ Bit32u fatDrive::getClustFirstSect(Bit32u clustNum) {
 }
 
 Bit32u fatDrive::getClusterValue(Bit32u clustNum) {
-	Bit32u fatoffset;
+	Bit32u fatoffset=0;
 	Bit32u fatsectnum;
 	Bit32u fatentoff;
-
-	Bit32u clustValue;
-
-
-	/* Load two sectors at once for FAT12 */
-	Bit8u sectbuffer[1024];
+	Bit32u clustValue=0;
 
 	switch(fattype) {
 		case FAT12:
@@ -265,12 +297,16 @@ Bit32u fatDrive::getClusterValue(Bit32u clustNum) {
 	fatsectnum = bootbuffer.reservedsectors + (fatoffset / bootbuffer.bytespersector) + partSectOff;
 	fatentoff = fatoffset % bootbuffer.bytespersector;
 
-	loadedDisk->Read_AbsoluteSector(fatsectnum, &sectbuffer[0]);
-	loadedDisk->Read_AbsoluteSector(fatsectnum+1, &sectbuffer[512]);
+	if(curFatSect != fatsectnum) {
+		/* Load two sectors at once for FAT12 */
+		loadedDisk->Read_AbsoluteSector(fatsectnum, &fatSectBuffer[0]);
+		loadedDisk->Read_AbsoluteSector(fatsectnum+1, &fatSectBuffer[512]);
+		curFatSect = fatsectnum;
+	}
 
 	switch(fattype) {
 		case FAT12:
-			clustValue = *((Bit16u *)&sectbuffer[fatentoff]);
+			clustValue = *((Bit16u *)&fatSectBuffer[fatentoff]);
 			if(clustNum & 0x1) {
 				clustValue >>= 4;
 			} else {
@@ -278,10 +314,10 @@ Bit32u fatDrive::getClusterValue(Bit32u clustNum) {
 			}
 			break;
 		case FAT16:
-			clustValue = *((Bit16u *)&sectbuffer[fatentoff]);
+			clustValue = *((Bit16u *)&fatSectBuffer[fatentoff]);
 			break;
 		case FAT32:
-			clustValue = *((Bit32u *)&sectbuffer[fatentoff]);
+			clustValue = *((Bit32u *)&fatSectBuffer[fatentoff]);
 			break;
 	}
 
@@ -289,14 +325,9 @@ Bit32u fatDrive::getClusterValue(Bit32u clustNum) {
 }
 
 void fatDrive::setClusterValue(Bit32u clustNum, Bit32u clustValue) {
-	Bit32u fatoffset;
+	Bit32u fatoffset=0;
 	Bit32u fatsectnum;
 	Bit32u fatentoff;
-	Bit32u tmpValue;
-
-
-	/* Load two sectors at once for FAT12 */
-	Bit8u sectbuffer[1024];
 
 	switch(fattype) {
 		case FAT12:
@@ -312,50 +343,52 @@ void fatDrive::setClusterValue(Bit32u clustNum, Bit32u clustValue) {
 	fatsectnum = bootbuffer.reservedsectors + (fatoffset / bootbuffer.bytespersector) + partSectOff;
 	fatentoff = fatoffset % bootbuffer.bytespersector;
 
-	loadedDisk->Read_AbsoluteSector(fatsectnum, &sectbuffer[0]);
-	loadedDisk->Read_AbsoluteSector(fatsectnum+1, &sectbuffer[512]);
+	if(curFatSect != fatsectnum) {
+		/* Load two sectors at once for FAT12 */
+		loadedDisk->Read_AbsoluteSector(fatsectnum, &fatSectBuffer[0]);
+		loadedDisk->Read_AbsoluteSector(fatsectnum+1, &fatSectBuffer[512]);
+		curFatSect = fatsectnum;
+	}
 
 	switch(fattype) {
-		case FAT12:
-			tmpValue = *((Bit16u *)&sectbuffer[fatentoff]);
+		case FAT12: {
+			Bit16u tmpValue = *((Bit16u *)&fatSectBuffer[fatentoff]);
 			if(clustNum & 0x1) {
 				clustValue &= 0xfff;
 				clustValue <<= 4;
 				tmpValue &= 0xf;
-				tmpValue |= clustValue;
+				tmpValue |= (Bit16u)clustValue;
 
 			} else {
 				clustValue &= 0xfff;
 				tmpValue &= 0xf000;
-				tmpValue |= clustValue;
+				tmpValue |= (Bit16u)clustValue;
 			}
-			*((Bit16u *)&sectbuffer[fatentoff]) = tmpValue;
+			*((Bit16u *)&fatSectBuffer[fatentoff]) = tmpValue;
 			break;
+			}
 		case FAT16:
-			*((Bit16u *)&sectbuffer[fatentoff]) = clustValue;
+			*((Bit16u *)&fatSectBuffer[fatentoff]) = (Bit16u)clustValue;
 			break;
 		case FAT32:
-			*((Bit32u *)&sectbuffer[fatentoff]) = clustValue;
+			*((Bit32u *)&fatSectBuffer[fatentoff]) = clustValue;
 			break;
 	}
 	int fc;
 	for(fc=0;fc<bootbuffer.fatcopies;fc++) {
-		loadedDisk->Write_AbsoluteSector(fatsectnum + (fc * bootbuffer.sectorsperfat), &sectbuffer[0]);
-		loadedDisk->Write_AbsoluteSector(fatsectnum+1+(fc * bootbuffer.sectorsperfat), &sectbuffer[512]);
+		loadedDisk->Write_AbsoluteSector(fatsectnum + (fc * bootbuffer.sectorsperfat), &fatSectBuffer[0]);
+		loadedDisk->Write_AbsoluteSector(fatsectnum+1+(fc * bootbuffer.sectorsperfat), &fatSectBuffer[512]);
 	}
 }
 
 bool fatDrive::getEntryName(char *fullname, char *entname) {
-	Bit16u len = strlen(fullname);
 	char dirtoken[DOS_PATHLENGTH];
-	Bit32u currentClust = 0;
 
-	direntry foundEntry;
 	char * findDir;
 	char * findFile;
 	strcpy(dirtoken,fullname);
 
-	LOG_MSG("Testing for filename %s", fullname);
+	//LOG_MSG("Testing for filename %s", fullname);
 	findDir = strtok(dirtoken,"\\");
 	findFile = findDir;
 	while(findDir != NULL) {
@@ -367,7 +400,7 @@ bool fatDrive::getEntryName(char *fullname, char *entname) {
 }
 
 bool fatDrive::getFileDirEntry(char * filename, direntry * useEntry, Bit32u * dirClust, Bit32u * subEntry) {
-	Bit16u len = strlen(filename);
+	Bit32u len = (Bit32u)strlen(filename);
 	char dirtoken[DOS_PATHLENGTH];
 	Bit32u currentClust = 0;
 
@@ -378,7 +411,7 @@ bool fatDrive::getFileDirEntry(char * filename, direntry * useEntry, Bit32u * di
 
 	/* Skip if testing in root directory */
 	if ((len>0) && (filename[len-1]!='\\')) {
-		LOG_MSG("Testing for filename %s", filename);
+		//LOG_MSG("Testing for filename %s", filename);
 		findDir = strtok(dirtoken,"\\");
 		findFile = findDir;
 		while(findDir != NULL) {
@@ -387,6 +420,13 @@ bool fatDrive::getFileDirEntry(char * filename, direntry * useEntry, Bit32u * di
 			
 			findFile = findDir;
 			if(!FindNextInternal(currentClust, *imgDTA, &foundEntry)) break;
+			else {
+				//Found something. See if it's a directory (findfirst always finds regular files)
+				char find_name[DOS_NAMELENGTH_ASCII];Bit16u find_date,find_time;Bit32u find_size;Bit8u find_attr;
+				imgDTA->GetResult(find_name,find_size,find_date,find_time,find_attr);
+				if(!(find_attr & DOS_ATTR_DIRECTORY)) break;
+			}
+
 			currentClust = foundEntry.loFirstClust;
 			findDir = strtok(NULL,"\\");
 		}
@@ -395,21 +435,18 @@ bool fatDrive::getFileDirEntry(char * filename, direntry * useEntry, Bit32u * di
 	}
 
 	/* Search found directory for our file */
-	imgDTA->SetupSearch(0,0x5,findFile);
+	imgDTA->SetupSearch(0,0x7,findFile);
 	imgDTA->SetDirID(0);
 	if(!FindNextInternal(currentClust, *imgDTA, &foundEntry)) return false;
-	
+
 	memcpy(useEntry, &foundEntry, sizeof(direntry));
 	*dirClust = (Bit32u)currentClust;
 	*subEntry = ((Bit32u)imgDTA->GetDirID()-1);
-	
-
-
 	return true;
 }
 
 bool fatDrive::getDirClustNum(char *dir, Bit32u *clustNum, bool parDir) {
-	Bit16u len = strlen(dir);
+	Bit32u len = (Bit32u)strlen(dir);
 	char dirtoken[DOS_PATHLENGTH];
 	Bit32u currentClust = 0;
 	direntry foundEntry;
@@ -418,30 +455,30 @@ bool fatDrive::getDirClustNum(char *dir, Bit32u *clustNum, bool parDir) {
 
 	/* Skip if testing for root directory */
 	if ((len>0) && (dir[len-1]!='\\')) {
-		LOG_MSG("Testing for dir %s", dir);
+		//LOG_MSG("Testing for dir %s", dir);
 		findDir = strtok(dirtoken,"\\");
 		while(findDir != NULL) {
 			imgDTA->SetupSearch(0,DOS_ATTR_DIRECTORY,findDir);
 			imgDTA->SetDirID(0);
 			findDir = strtok(NULL,"\\");
-			if(!parDir) {
-			if(!FindNextInternal(currentClust, *imgDTA, &foundEntry)) return false;
+			if(parDir && (findDir == NULL)) break;
+
+			char find_name[DOS_NAMELENGTH_ASCII];Bit16u find_date,find_time;Bit32u find_size;Bit8u find_attr;
+			if(!FindNextInternal(currentClust, *imgDTA, &foundEntry)) {
+				return false;
 			} else {
-				if(findDir == NULL) break;
-				if(!FindNextInternal(currentClust, *imgDTA, &foundEntry)) return false;
+				imgDTA->GetResult(find_name,find_size,find_date,find_time,find_attr);
+				if(!(find_attr &DOS_ATTR_DIRECTORY)) return false;
 			}
 			currentClust = foundEntry.loFirstClust;
-			
+
 		}
 		*clustNum = currentClust;
-		return true;
 	} else {
 		/* Set to root directory */
 		*clustNum = 0;
-		return true;
 	}
-	return false;
-
+	return true;
 }
 
 Bit32u fatDrive::getSectorSize(void) {
@@ -449,7 +486,7 @@ Bit32u fatDrive::getSectorSize(void) {
 }
 
 Bit32u fatDrive::getAbsoluteSectFromBytePos(Bit32u startClustNum, Bit32u bytePos) {
-	return getAbsoluteSectFromChain(startClustNum, bytePos / bootbuffer.bytespersector);
+	return  getAbsoluteSectFromChain(startClustNum, bytePos / bootbuffer.bytespersector);
 }
 
 Bit32u fatDrive::getAbsoluteSectFromChain(Bit32u startClustNum, Bit32u logicalSector) {
@@ -473,13 +510,14 @@ Bit32u fatDrive::getAbsoluteSectFromChain(Bit32u startClustNum, Bit32u logicalSe
 				if(testvalue >= 0xfffffff8) isEOF = true;
 				break;
 		}
-		if((isEOF) && (skipClust>1)) {
+		if((isEOF) && (skipClust>=1)) {
 			//LOG_MSG("End of cluster chain reached before end of logical sector seek!");
 			return 0;
 		}
 		currentClust = testvalue;
 		--skipClust;
 	}
+
 	return (getClustFirstSect(currentClust) + sectClust);
 }
 
@@ -542,10 +580,6 @@ Bit32u fatDrive::appendCluster(Bit32u startCluster) {
 	zeroOutCluster(newClust);
 
 	return newClust;
-
-
-
-
 }
 
 bool fatDrive::allocateCluster(Bit32u useCluster, Bit32u prevCluster) {
@@ -559,6 +593,7 @@ bool fatDrive::allocateCluster(Bit32u useCluster, Bit32u prevCluster) {
 
 		/* Point cluster to new cluster in chain */
 		setClusterValue(prevCluster, useCluster);
+		//LOG_MSG("Chaining cluser %d to %d", prevCluster, useCluster);
 	} 
 
 	switch(fattype) {
@@ -573,10 +608,10 @@ bool fatDrive::allocateCluster(Bit32u useCluster, Bit32u prevCluster) {
 			break;
 	}
 	return true;
-
 }
 
 fatDrive::fatDrive(const char *sysFilename, Bit32u bytesector, Bit32u cylsector, Bit32u headscyl, Bit32u cylinders, Bit32u startSector) {
+	created_succesfully = true;
 	FILE *diskfile;
 	Bit32u filesize;
 	struct partTable mbrData;
@@ -588,14 +623,14 @@ fatDrive::fatDrive(const char *sysFilename, Bit32u bytesector, Bit32u cylsector,
 	}
 
 	diskfile = fopen(sysFilename, "rb+");
-	if(!diskfile) return;
+	if(!diskfile) {created_succesfully = false;return;}
 	fseek(diskfile, 0L, SEEK_END);
 	filesize = (Bit32u)ftell(diskfile) / 1024L;
 
-    /* Load disk image */
+	/* Load disk image */
 	loadedDisk = new imageDisk(diskfile, (Bit8u *)sysFilename, filesize, (filesize > 2880));
 	if(!loadedDisk) {
-		delete this;
+		created_succesfully = false;
 		return;
 	}
 
@@ -604,6 +639,9 @@ fatDrive::fatDrive(const char *sysFilename, Bit32u bytesector, Bit32u cylsector,
 		loadedDisk->Set_Geometry(headscyl, cylinders,cylsector, bytesector);
 
 		loadedDisk->Read_Sector(0,0,1,&mbrData);
+
+		if(mbrData.magic1!= 0x55 ||	mbrData.magic2!= 0xaa) LOG_MSG("Possibly invalid partition table in disk image.");
+
 		startSector = 63;
 		int m;
 		for(m=0;m<4;m++) {
@@ -615,6 +653,8 @@ fatDrive::fatDrive(const char *sysFilename, Bit32u bytesector, Bit32u cylsector,
 			}
 		}
 
+		if(m==4) LOG_MSG("No good partiton found in image.");
+
 		partSectOff = startSector;
 	} else {
 		/* Floppy disks don't have partitions */
@@ -624,13 +664,12 @@ fatDrive::fatDrive(const char *sysFilename, Bit32u bytesector, Bit32u cylsector,
 	loadedDisk->Read_AbsoluteSector(0+partSectOff,&bootbuffer);
 	if ((bootbuffer.magic1 != 0x55) || (bootbuffer.magic2 != 0xaa)) {
 		/* Not a FAT filesystem */
-		delete this;
-		return;
+		LOG_MSG("Loaded image has no valid magicnumbers at the end!");
 	}
 
 	if(!bootbuffer.sectorsperfat) {
 		/* FAT32 not implemented yet */
-		delete this;
+		created_succesfully = false;
 		return;
 	}
 
@@ -667,22 +706,29 @@ fatDrive::fatDrive(const char *sysFilename, Bit32u bytesector, Bit32u cylsector,
 	/* There is no cluster 0, this means we are in the root directory */
 	cwdDirCluster = 0;
 
-
-
-
+	memset(fatSectBuffer,0,1024);
+	curFatSect = 0xffffffff;
 }
 
 bool fatDrive::AllocationInfo(Bit16u *_bytes_sector, Bit8u *_sectors_cluster, Bit16u *_total_clusters, Bit16u *_free_clusters) {
 	Bit32u hs, cy, sect,sectsize;
 	Bit32u countFree = 0;
-	int i;
+	Bit32u i;
 	
 	loadedDisk->Get_Geometry(&hs, &cy, &sect, &sectsize);
 	*_bytes_sector = (Bit16u)sectsize;
 	*_sectors_cluster = bootbuffer.sectorspercluster;
-	*_total_clusters = CountOfClusters;
+	if (CountOfClusters<65536) *_total_clusters = (Bit16u)CountOfClusters;
+	else {
+		// maybe some special handling needed for fat32
+		*_total_clusters = 65535;
+	}
 	for(i=0;i<CountOfClusters;i++) if(!getClusterValue(i+2)) countFree++;
-	*_free_clusters = countFree;
+	if (countFree<65536) *_free_clusters = (Bit16u)countFree;
+	else {
+		// maybe some special handling needed for fat32
+		*_free_clusters = 65535;
+	}
 	
 	return true;
 }
@@ -695,12 +741,16 @@ Bit32u fatDrive::getFirstFreeClust(void) {
 
 	/* No free cluster found */
 	return 0;
-	
 }
 
-bool fatDrive::isRemote(void) { return false; }
+bool fatDrive::isRemote(void) {	return false; }
 bool fatDrive::isRemovable(void) { return false; }
 
+Bits fatDrive::UnMount(void) {
+	delete this;
+	return 0;
+}
+
 Bit8u fatDrive::GetMediaByte(void) { return loadedDisk->GetBiosType(); }
 
 bool fatDrive::FileCreate(DOS_File **file, char *name, Bit16u attributes) {
@@ -710,28 +760,34 @@ bool fatDrive::FileCreate(DOS_File **file, char *name, Bit16u attributes) {
 	char pathName[11];
 
 	/* Check if file already exists */
-	if(getFileDirEntry(name, &fileEntry, &dirClust, &subEntry)) return false;
+	if(getFileDirEntry(name, &fileEntry, &dirClust, &subEntry)) {
+		/* Truncate file */
+		fileEntry.entrysize=0;
+		directoryChange(dirClust, &fileEntry, subEntry);
+	} else {
+		/* Can we even get the name of the file itself? */
+		if(!getEntryName(name, &dirName[0])) return false;
+		convToDirFile(&dirName[0], &pathName[0]);
 
-	/* Can we even get the name of the file itself? */
-	if(!getEntryName(name, &dirName[0])) return false;
-	convToDirFile(&dirName[0], &pathName[0]);
+		/* Can we find the base directory? */
+		if(!getDirClustNum(name, &dirClust, true)) return false;
+		memset(&fileEntry, 0, sizeof(direntry));
+		memcpy(&fileEntry.entryname, &pathName[0], 11);
+		fileEntry.attrib = (Bit8u)(attributes & 0xff);
+		addDirectoryEntry(dirClust, fileEntry);
 
-	/* Can we find the base directory? */
-	if(!getDirClustNum(name, &dirClust, true)) return false;
-	memset(&fileEntry, 0, sizeof(direntry));
-	memcpy(&fileEntry.entryname, &pathName[0], 11);
-	fileEntry.attrib = attributes;
-	addDirectoryEntry(dirClust, fileEntry);
-
-	/* Check if file exists now */
-	if(!getFileDirEntry(name, &fileEntry, &dirClust, &subEntry)) return false;
+		/* Check if file exists now */
+		if(!getFileDirEntry(name, &fileEntry, &dirClust, &subEntry)) return false;
+	}
 
 	/* Empty file created, now lets open it */
 	/* TODO: check for read-only flag and requested write access */
 	*file = new fatFile(name, fileEntry.loFirstClust, fileEntry.entrysize, this);
 	((fatFile *)(*file))->dirCluster = dirClust;
 	((fatFile *)(*file))->dirIndex = subEntry;
-	
+	/* Maybe modTime and date should be used ? (crt matches findnext) */
+	((fatFile *)(*file))->time = fileEntry.crtTime;
+	((fatFile *)(*file))->date = fileEntry.crtDate;
 	return true;
 }
 
@@ -742,7 +798,7 @@ bool fatDrive::FileExists(const char *name) {
 	return true;
 }
 
-bool fatDrive::FileOpen(DOS_File **file, char *name, Bit32u flags) {
+bool fatDrive::FileOpen(DOS_File **file, char *name, Bit32u /*flags*/) {
 	direntry fileEntry;
 	Bit32u dirClust, subEntry;
 	if(!getFileDirEntry(name, &fileEntry, &dirClust, &subEntry)) return false;
@@ -750,11 +806,13 @@ bool fatDrive::FileOpen(DOS_File **file, char *name, Bit32u flags) {
 	*file = new fatFile(name, fileEntry.loFirstClust, fileEntry.entrysize, this);
 	((fatFile *)(*file))->dirCluster = dirClust;
 	((fatFile *)(*file))->dirIndex = subEntry;
-	
+	/* Maybe modTime and date should be used ? (crt matches findnext) */
+	((fatFile *)(*file))->time = fileEntry.crtTime;
+	((fatFile *)(*file))->date = fileEntry.crtDate;
 	return true;
 }
 
-bool fatDrive::FileStat(const char *name, FileStat_Block *const stat_block) {
+bool fatDrive::FileStat(const char * /*name*/, FileStat_Block *const /*stat_block*/) {
 	/* TODO: Stub */
 	return false;
 }
@@ -773,14 +831,26 @@ bool fatDrive::FileUnlink(char * name) {
 	return true;
 }
 
-bool fatDrive::FindFirst(char *_dir, DOS_DTA &dta,bool fcb_findfirst) {
+bool fatDrive::FindFirst(char *_dir, DOS_DTA &dta,bool /*fcb_findfirst*/) {
 	direntry dummyClust;
 	Bit8u attr;char pattern[DOS_NAMELENGTH_ASCII];
 	dta.GetSearchParams(attr,pattern);
+	if(attr==DOS_ATTR_VOLUME) {
+		if (strcmp(GetLabel(), "") == 0 ) {
+			DOS_SetError(DOSERR_NO_MORE_FILES);
+			return false;
+		}
+		dta.SetResult(GetLabel(),0,0,0,DOS_ATTR_VOLUME);
+		return true;
+	}
 	if(attr & DOS_ATTR_VOLUME) //check for root dir or fcb_findfirst
 		LOG(LOG_DOSMISC,LOG_WARN)("findfirst for volumelabel used on fatDrive. Unhandled!!!!!");
-	if(!getDirClustNum(_dir, &cwdDirCluster, false)) return false;
+	if(!getDirClustNum(_dir, &cwdDirCluster, false)) {
+		DOS_SetError(DOSERR_PATH_NOT_FOUND);
+		return false;
+	}
 	dta.SetDirID(0);
+	dta.SetDirIDCluster((Bit16u)(cwdDirCluster&0xffff));
 	return FindNextInternal(cwdDirCluster, dta, &dummyClust);
 }
 
@@ -828,7 +898,10 @@ nextfile:
 	} else {
 		tmpsector = getAbsoluteSectFromChain(dirClustNumber, logentsector);
 		/* A zero sector number can't happen */
-		if(tmpsector == 0) return false;
+		if(tmpsector == 0) {
+			DOS_SetError(DOSERR_NO_MORE_FILES);
+			return false;
+		}
 		loadedDisk->Read_AbsoluteSector(tmpsector,sectbuf);
 	}
 	dirPos++;
@@ -838,7 +911,10 @@ nextfile:
 	if (sectbuf[entryoffset].entryname[0] == 0xe5) goto nextfile;
 
 	/* End of directory list */
-	if (sectbuf[entryoffset].entryname[0] == 0x00) return false;
+	if (sectbuf[entryoffset].entryname[0] == 0x00) {
+		DOS_SetError(DOSERR_NO_MORE_FILES);
+		return false;
+	}
 
 	memset(find_name,0,DOS_NAMELENGTH_ASCII);
 	memset(extension,0,4);
@@ -846,42 +922,65 @@ nextfile:
 	memcpy(extension,&sectbuf[entryoffset].entryname[8],3);
 	trimString(&find_name[0]);
 	trimString(&extension[0]);
-	if(!(sectbuf[entryoffset].attrib & DOS_ATTR_DIRECTORY)) { 
+	if(!(sectbuf[entryoffset].attrib & DOS_ATTR_DIRECTORY) || extension[0]!=0) { 
 		strcat(find_name, ".");
 		strcat(find_name, extension);
 	}
 
-	if((attrs & (sectbuf[entryoffset].attrib | 0x21)) == 0)  goto nextfile;
+	/* Ignore files with volume label. FindFirst should search for those. (return the first one found) */
+	if(sectbuf[entryoffset].attrib & 0x8) goto nextfile;
+   
+	/* Always find ARCHIVES even if bit is not set  Perhaps test is not the best test */
+	if(~attrs & sectbuf[entryoffset].attrib & (DOS_ATTR_DIRECTORY | DOS_ATTR_HIDDEN | DOS_ATTR_SYSTEM) )  goto nextfile;
 	if(!WildFileCmp(find_name,srch_pattern)) goto nextfile;
 
 	dta.SetResult(find_name, sectbuf[entryoffset].entrysize, sectbuf[entryoffset].crtDate, sectbuf[entryoffset].crtTime, sectbuf[entryoffset].attrib);
 	memcpy(foundEntry, &sectbuf[entryoffset], sizeof(direntry));
 
 	return true;
-
 }
 
 bool fatDrive::FindNext(DOS_DTA &dta) {
 	direntry dummyClust;
 
-	return FindNextInternal(cwdDirCluster, dta, &dummyClust);
+	return FindNextInternal(dta.GetDirIDCluster(), dta, &dummyClust);
 }
 
 bool fatDrive::GetFileAttr(char *name, Bit16u *attr) {
-	/* TODO: Stub */
-	return false;
+	direntry fileEntry;
+	Bit32u dirClust, subEntry;
+	if(!getFileDirEntry(name, &fileEntry, &dirClust, &subEntry)) {
+		char dirName[DOS_NAMELENGTH_ASCII];
+		char pathName[11];
+
+		/* Can we even get the name of the directory itself? */
+		if(!getEntryName(name, &dirName[0])) return false;
+		convToDirFile(&dirName[0], &pathName[0]);
+
+		/* Get parent directory starting cluster */
+		if(!getDirClustNum(name, &dirClust, true)) return false;
+
+		/* Find directory entry in parent directory */
+		Bit32s fileidx = 2;
+		if (dirClust==0) fileidx = 0;	// root directory
+		while(directoryBrowse(dirClust, &fileEntry, fileidx)) {
+			if(memcmp(&fileEntry.entryname, &pathName[0], 11) == 0) {
+				*attr=fileEntry.attrib;
+				return true;
+			}
+			fileidx++;
+		}
+		return false;
+	} else *attr=fileEntry.attrib;
+	return true;
 }
 
 bool fatDrive::directoryBrowse(Bit32u dirClustNumber, direntry *useEntry, Bit32s entNum) {
-	direntry sectbuf[16]; /* 16 directory entries per sector */
-	Bit32u logentsector; /* Logical entry sector */
-	Bit32u entryoffset;  /* Index offset within sector */
+	direntry sectbuf[16];	/* 16 directory entries per sector */
+	Bit32u logentsector;	/* Logical entry sector */
+	Bit32u entryoffset = 0;	/* Index offset within sector */
 	Bit32u tmpsector;
-	Bit8u attrs;
 	Bit16u dirPos = 0;
-	char srch_pattern[DOS_NAMELENGTH_ASCII];
-	char find_name[DOS_NAMELENGTH_ASCII];
-	char extension[4];
 	
 	while(entNum>=0) {
 		
@@ -911,15 +1010,11 @@ bool fatDrive::directoryBrowse(Bit32u dirClustNumber, direntry *useEntry, Bit32s
 }
 
 bool fatDrive::directoryChange(Bit32u dirClustNumber, direntry *useEntry, Bit32s entNum) {
-	direntry sectbuf[16]; /* 16 directory entries per sector */
-	Bit32u logentsector; /* Logical entry sector */
-	Bit32u entryoffset;  /* Index offset within sector */
+	direntry sectbuf[16];	/* 16 directory entries per sector */
+	Bit32u logentsector;	/* Logical entry sector */
+	Bit32u entryoffset = 0;	/* Index offset within sector */
 	Bit32u tmpsector = 0;
-	Bit8u attrs;
 	Bit16u dirPos = 0;
-	char srch_pattern[DOS_NAMELENGTH_ASCII];
-	char find_name[DOS_NAMELENGTH_ASCII];
-	char extension[4];
 	
 	while(entNum>=0) {
 		
@@ -957,13 +1052,9 @@ bool fatDrive::addDirectoryEntry(Bit32u dirClustNumber, direntry useEntry) {
 	Bit32u logentsector; /* Logical entry sector */
 	Bit32u entryoffset;  /* Index offset within sector */
 	Bit32u tmpsector;
-	Bit8u attrs;
 	Bit16u dirPos = 0;
-	char srch_pattern[DOS_NAMELENGTH_ASCII];
-	char find_name[DOS_NAMELENGTH_ASCII];
-	char extension[4];
 	
-	while(true) {
+	for(;;) {
 		
 		logentsector = dirPos / 16;
 		entryoffset = dirPos % 16;
@@ -991,12 +1082,11 @@ bool fatDrive::addDirectoryEntry(Bit32u dirClustNumber, direntry useEntry) {
 		if ((sectbuf[entryoffset].entryname[0] == 0xe5) || (sectbuf[entryoffset].entryname[0] == 0x00)) {
 			sectbuf[entryoffset] = useEntry;
 			loadedDisk->Write_AbsoluteSector(tmpsector,sectbuf);
-			return true;
+			break;
 		}
 	}
 
-
-	return false;
+	return true;
 }
 
 void fatDrive::zeroOutCluster(Bit32u clustNumber) {
@@ -1008,7 +1098,6 @@ void fatDrive::zeroOutCluster(Bit32u clustNumber) {
 	for(i=0;i<bootbuffer.sectorspercluster;i++) {
 		loadedDisk->Write_AbsoluteSector(getAbsoluteSectFromChain(clustNumber,i), &secBuffer[0]);
 	}
-
 }
 
 bool fatDrive::MakeDir(char *dir) {
@@ -1061,7 +1150,6 @@ bool fatDrive::MakeDir(char *dir) {
 	addDirectoryEntry(dummyClust, tmpentry);
 
 	return true;
-
 }
 
 bool fatDrive::RemoveDir(char *dir) {
@@ -1097,7 +1185,8 @@ bool fatDrive::RemoveDir(char *dir) {
 	if(filecount > 0) return false;
 
 	/* Find directory entry in parent directory */
-	fileidx = 2;
+	if (dirClust==0) fileidx = 0;	// root directory
+	else fileidx = 2;
 	bool found = false;
 	while(directoryBrowse(dirClust, &tmpentry, fileidx)) {
 		if(memcmp(&tmpentry.entryname, &pathName[0], 11) == 0) {
@@ -1116,8 +1205,7 @@ bool fatDrive::RemoveDir(char *dir) {
 	return true;
 }
 
-bool fatDrive::Rename(char *oldname, char*newname) {
-	
+bool fatDrive::Rename(char * /*oldname*/, char * /*newname*/) {
 	return false;
 }
 
