@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2006  The DOSBox Team
+ *  Copyright (C) 2002-2007  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: memory.cpp,v 1.44 2006/02/09 11:47:49 qbix79 Exp $ */
+/* $Id: memory.cpp,v 1.48 2007/01/14 18:44:01 c2woody Exp $ */
 
 #include "dosbox.h"
 #include "mem.h"
@@ -291,6 +291,10 @@ MemHandle MEM_AllocatePages(Bitu pages,bool sequence) {
 	return ret;
 }
 
+MemHandle MEM_GetNextFreePage(void) {
+	return (MemHandle)BestMatch(1);
+}
+
 void MEM_ReleasePages(MemHandle handle) {
 	while (handle>0) {
 		MemHandle next=memory.mhandles[handle];
@@ -300,7 +304,7 @@ void MEM_ReleasePages(MemHandle handle) {
 }
 
 bool MEM_ReAllocatePages(MemHandle & handle,Bitu pages,bool sequence) {
-	if (handle<0) {
+	if (handle<=0) {
 		if (!pages) return true;
 		handle=MEM_AllocatePages(pages,sequence);
 		return (handle>0);
@@ -494,6 +498,13 @@ static Bitu read_p92(Bitu port,Bitu iolen) {
 	return memory.a20.controlport | (memory.a20.enabled ? 0x02 : 0);
 }
 
+void PreparePCJRCartRom(void) {
+	/* Setup rom at 0xd0000-0xe0000 */
+	for (Bitu ct=0xd0;ct<0xe0;ct++) {
+		memory.phandlers[ct] = &rom_page_handler;
+	}
+}
+
 HostPt GetMemBase(void) { return MemBase; }
 
 class MEMORY:public Module_base{
