@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2006  The DOSBox Team
+ *  Copyright (C) 2002-2007  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: mixer.cpp,v 1.39 2006/03/27 19:41:55 qbix79 Exp $ */
+/* $Id: mixer.cpp,v 1.44 2007/01/08 19:45:40 qbix79 Exp $ */
 
 /* 
 	Remove the sdl code from here and have it handeld in the sdlmain.
@@ -174,11 +174,13 @@ void MixerChannel::AddSilence(void) {
 	}
 }
 
-template<bool _8bits,bool stereo>
+template<bool _8bits,bool stereo,bool signeddata>
 INLINE void MixerChannel::AddSamples(Bitu len,void * data) {
 	Bits diff[2];
 	Bit8u * data8=(Bit8u*)data;
+	Bit8s * data8s=(Bit8s*)data;
 	Bit16s * data16=(Bit16s*)data;
+	Bit16u * data16u=(Bit16u*)data;
 	Bitu mixpos=mixer.pos+done;
 	freq_index&=MIXER_REMAIN;
 	Bitu pos=0;Bitu new_pos;
@@ -193,18 +195,36 @@ INLINE void MixerChannel::AddSamples(Bitu len,void * data) {
 thestart:
 			if (pos>=len) return;
 			if (_8bits) {
-				if (stereo) {
-					diff[0]=(((Bit8s)(data8[pos*2+0] ^ 0x80)) << 8)-last[0];
-					diff[1]=(((Bit8s)(data8[pos*2+1] ^ 0x80)) << 8)-last[1];
+				if (!signeddata) {
+					if (stereo) {
+						diff[0]=(((Bit8s)(data8[pos*2+0] ^ 0x80)) << 8)-last[0];
+						diff[1]=(((Bit8s)(data8[pos*2+1] ^ 0x80)) << 8)-last[1];
+					} else {
+						diff[0]=(((Bit8s)(data8[pos] ^ 0x80)) << 8)-last[0];
+					}
 				} else {
-					diff[0]=(((Bit8s)(data8[pos] ^ 0x80)) << 8)-last[0];
+					if (stereo) {
+						diff[0]=(data8s[pos*2+0] << 8)-last[0];
+						diff[1]=(data8s[pos*2+1] << 8)-last[1];
+					} else {
+						diff[0]=(data8s[pos] << 8)-last[0];
+					}
 				}
 			} else {
-				if (stereo) {
-					diff[0]=data16[pos*2+0]-last[0];
-					diff[1]=data16[pos*2+1]-last[1];
+				if (signeddata) {
+					if (stereo) {
+						diff[0]=data16[pos*2+0]-last[0];
+						diff[1]=data16[pos*2+1]-last[1];
+					} else {
+						diff[0]=data16[pos]-last[0];
+					}
 				} else {
-					diff[0]=data16[pos]-last[0];
+					if (stereo) {
+						diff[0]=(Bits)data16u[pos*2+0]-32768-last[0];
+						diff[1]=(Bits)data16u[pos*2+1]-32768-last[1];
+					} else {
+						diff[0]=(Bits)data16u[pos]-32768-last[0];
+					}
 				}
 			}
 		}
@@ -248,16 +268,28 @@ void MixerChannel::AddStretched(Bitu len,Bit16s * data) {
 };
 
 void MixerChannel::AddSamples_m8(Bitu len,Bit8u * data) {
-	AddSamples<true,false>(len,data);
+	AddSamples<true,false,false>(len,data);
 }
 void MixerChannel::AddSamples_s8(Bitu len,Bit8u * data) {
-	AddSamples<true,true>(len,data);
+	AddSamples<true,true,false>(len,data);
+}
+void MixerChannel::AddSamples_m8s(Bitu len,Bit8s * data) {
+	AddSamples<true,false,true>(len,data);
+}
+void MixerChannel::AddSamples_s8s(Bitu len,Bit8s * data) {
+	AddSamples<true,true,true>(len,data);
 }
 void MixerChannel::AddSamples_m16(Bitu len,Bit16s * data) {
-	AddSamples<false,false>(len,data);
+	AddSamples<false,false,true>(len,data);
 }
 void MixerChannel::AddSamples_s16(Bitu len,Bit16s * data) {
-	AddSamples<false,true>(len,data);
+	AddSamples<false,true,true>(len,data);
+}
+void MixerChannel::AddSamples_m16u(Bitu len,Bit16u * data) {
+	AddSamples<false,false,false>(len,data);
+}
+void MixerChannel::AddSamples_s16u(Bitu len,Bit16u * data) {
+	AddSamples<false,true,false>(len,data);
 }
 
 void MixerChannel::FillUp(void) {
