@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2006  The DOSBox Team
+ *  Copyright (C) 2002-2007  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: bios.cpp,v 1.57 2006/02/09 11:47:55 qbix79 Exp $ */
+/* $Id: bios.cpp,v 1.66 2007/02/22 08:33:15 qbix79 Exp $ */
 
 #include "dosbox.h"
 #include "mem.h"
@@ -29,6 +29,7 @@
 #include "joystick.h"
 #include "mouse.h"
 #include "setup.h"
+#include "serialport.h"
 
 
 /* if mem_systems 0 then size_extended is reported as the real size else 
@@ -62,7 +63,7 @@ static Bitu INT70_Handler(void) {
 	return 0;
 }
 
-CALLBACK_HandlerObject* tandy_DAC_callback;
+CALLBACK_HandlerObject* tandy_DAC_callback[2];
 static struct {
 	Bit16u port;
 	Bit8u irq;
@@ -137,9 +138,9 @@ static void Tandy_SetupTransfer(PhysPt bufpt,bool isplayback) {
 
 	/* revector IRQ-handler if necessary */
 	RealPt current_irq=RealGetVec(tandy_sb.irq+8);
-	if (current_irq!=tandy_DAC_callback->Get_RealPointer()) {
+	if (current_irq!=tandy_DAC_callback[0]->Get_RealPointer()) {
 		real_writed(0x40,0xd6,current_irq);
-		RealSetVec(tandy_sb.irq+8,tandy_DAC_callback->Get_RealPointer());
+		RealSetVec(tandy_sb.irq+8,tandy_DAC_callback[0]->Get_RealPointer());
 	}
 
 	IO_Write(tandy_sb.port+0xc,0xd0);	/* stop DMA transfer */
@@ -216,12 +217,8 @@ static Bitu IRQ_TandyDAC(void) {
 		IO_Read(tandy_sb.port+0xe);
 
 		/* issue BIOS tandy sound device busy callout */
-		Bit16u oldax=reg_ax;
-		reg_ax=0x91fb;
-		CALLBACK_RunRealInt(0x15);
-		reg_ax = oldax;
-
-		IO_Write(0x20,0x20);
+		SegSet16(cs, RealSeg(tandy_DAC_callback[1]->Get_RealPointer()));
+		reg_ip = RealOff(tandy_DAC_callback[1]->Get_RealPointer());
 	}
 	return CBRET_NONE;
 }
@@ -332,16 +329,6 @@ static Bitu INT8_Handler(void) {
 	if (val) mem_writeb(BIOS_DISK_MOTOR_TIMEOUT,val-1);
 	/* and running drive */
 	mem_writeb(BIOS_DRIVE_RUNNING,mem_readb(BIOS_DRIVE_RUNNING) & 0xF0);
-	// Save ds,dx,ax
-	Bit16u oldds = SegValue(ds);
-	Bit16u olddx = reg_dx;
-	Bit16u oldax = reg_ax;
-	// run int 1c	
-	CALLBACK_RunRealInt(0x1c);
-	// restore old values
-	SegSet16(ds,oldds);
-	reg_dx = olddx;
-	reg_ax = oldax;
 	return CBRET_NONE;
 };
 
@@ -375,10 +362,15 @@ static Bitu INT17_Handler(void) {
 
 static Bitu INT14_Handler(void)
 {
+	if(reg_ah > 0x3 || reg_dx > 0x3) {	// 0-3 serial port functions
+										// and no more than 4 serial ports
+		LOG_MSG("BIOS INT14: Unhandled call AH=%2X DX=%4x",reg_dx);
+		return CBRET_NONE;
+	}
+	
 	Bit16u port = real_readw(0x40,reg_dx*2); // DX is always port number
-	if(reg_dx > 0x3 || port==0)	// no more than 4 serial ports
-	{
-		LOG_MSG("BIOS INT14: port %d does not exist.",reg_dx);
+	if(port==0)	{
+		LOG(LOG_BIOS,LOG_NORMAL)("BIOS INT14: port %d does not exist.",reg_dx);
 		return CBRET_NONE;
 	}
 	switch (reg_ah)
@@ -416,12 +408,11 @@ static Bitu INT14_Handler(void)
 			// disable interrupts
 			IO_WriteB(port+1, 0);
 			IO_ReadB(port+2);
-			// put RTS and DTR on
-			IO_WriteB(port+4,0x3);
 
 			// get result
 			reg_ah=IO_ReadB(port+5);
 			reg_al=IO_ReadB(port+6);
+			CALLBACK_SCF(false);
 		}
 		break;
 	case 0x01:	/* Write character */
@@ -431,126 +422,49 @@ static Bitu INT14_Handler(void)
 		// AH: line status
 		// AL: modem status
 		{
-			if((IO_ReadB(port+5)&&0x20)==0)
-			{
-				// TODO: should wait until they become empty->timeout
-				LOG_MSG("BIOS INT14: port %d: transmit register not empty.",reg_dx);
-				reg_ah = IO_ReadB(port+5)|0x80;
-				return CBRET_NONE;
+			if(serialports[reg_dx]) {
+				bool timeout;
+				// switch modem lines on
+				IO_WriteB(port+4,0x3);
+				timeout = !serialports[reg_dx]->Putchar(reg_al,true,true,
+					mem_readb(BIOS_COM1_TIMEOUT+reg_dx)*1000);
+				// get result
+				reg_ah=IO_ReadB(port+5);
+				if(timeout) reg_ah |= 0x80;
+				reg_al=IO_ReadB(port+6);
 			}
-			// transmit it
-			IO_WriteB(port,reg_al);
-			
-			if((IO_ReadB(port+5)&&0x60)==0)
-			{
-				// TODO: should wait until they become empty->timeout
-				LOG_MSG("BIOS INT14: port %d: transmit register not empty after write.",reg_dx);
-				reg_ah = IO_ReadB(port+5)|0x80;
-				return CBRET_NONE;
-			}
-			
-			// get result
-			reg_ah=IO_ReadB(port+5);
-			reg_al=IO_ReadB(port+6);
+			CALLBACK_SCF(false);
 		}
 		break;
 	
 	case 0x02:	/* Read character */
 		{
-			if((IO_ReadB(port+5)&0x1)!=0)
-			{
-				reg_al=IO_ReadB(port);
+			if(serialports[reg_dx]) {
+				bool timeout;
+				Bit8u buffer;
+				// switch modem lines on
+				IO_WriteB(port+4,0x3);
+				// wait for something
+				timeout = !serialports[reg_dx]->Getchar(&buffer,true, 
+					mem_readb(BIOS_COM1_TIMEOUT+reg_dx)*1000);
+
+				// RTS off
+				IO_WriteB(port+4,0x1);
+				// get result
+				reg_ah=IO_ReadB(port+5);
+				if(timeout) reg_ah |= 0x80;
+				else reg_al=buffer;
 			}
-			else
-			{
-				// TODO: should wait until timeout
-				LOG_MSG("BIOS INT14: port %d: nothing received.",reg_dx);
-				reg_ah = IO_ReadB(port+5)|0x80;
-				return CBRET_NONE;
-			}
-			reg_ah=IO_ReadB(port+5);
+			CALLBACK_SCF(false);
+			break;
 		}
-		break;
 	case 0x03: // get status
 		{
 			reg_ah=IO_ReadB(port+5);
-			//LOG_MSG("status reg_ah: %x",reg_ah);
 			reg_al=IO_ReadB(port+6);
+			CALLBACK_SCF(false);
 		}
-		break;
-	case 0x04:	// extended initialisation
-		// Parameters:
-		// AL: break
-		// BH: parity
-		// BL: stopbit
-		// CH: word length
-		// CL: baudrate
-		{
-			Bit8u lcr = 0;
-			
-			// baud rate
-			Bitu baudrate = 9600;
-			Bit16u baudresult;
-			Bitu rawbaud=reg_cl;
-			
-			if(rawbaud==0){ baudrate=110;}
-			else if (rawbaud==1){ baudrate=150;}
-			else if (rawbaud==2){ baudrate=300;}
-			else if (rawbaud==3){ baudrate=600;}
-			else if (rawbaud==4){ baudrate=1200;}
-			else if (rawbaud==5){ baudrate=2400;}
-			else if (rawbaud==6){ baudrate=4800;}
-			else if (rawbaud==7){ baudrate=9600;}
-			else if (rawbaud==8){ baudrate=19200;}
-
-			baudresult = (Bit16u)(115200 / baudrate);
-
-			IO_WriteB(port+3, 0x80);	// enable divider access
-			IO_WriteB(port,(Bit8u)baudresult&0xff);
-			IO_WriteB(port+1,(Bit8u)(baudresult>>8));
-			
-			// line configuration
-			// break
-			if(reg_al!=0) lcr=0x40;
-			// parity
-			if(reg_bh!=0)
-			{
-				if(reg_bh==1)lcr|=0x8;// odd
-				else if(reg_bh==2)lcr|=0x18;// even
-				else if(reg_bh==3)lcr|=0x28;// mark
-				else if(reg_bh==4)lcr|=0x38;// mark
-			}
-			// stopbit
-			if(reg_bl!=0)
-			{
-				lcr|=0x4;
-			}
-			// data length
-			lcr|=(reg_ch&0x3);
-			IO_WriteB(port+3,lcr);
-
-			reg_ah=IO_ReadB(port+5);
-			reg_al=IO_ReadB(port+6);
-		}
-		break;
-	case 0x05:	// modem control
-		{
-			if(reg_al==0)	// read MCR
-			{
-				reg_bl=IO_ReadB(port+4);
-			}
-			else if(reg_al==1)	// write MCR
-			{
-				IO_WriteB(port+4,reg_bl);
-			}
-			else LOG_MSG("BIOS INT14: port %d, function 5: invalid subfunction.",reg_dx);
-			reg_ah=IO_ReadB(port+5);
-			reg_al=IO_ReadB(port+6);
-		}
-		break;
-	default:
-		LOG_MSG("Unhandled INT 14 call %2X",reg_ah);
-		
+		break;		
 	}
 	return CBRET_NONE;
 }
@@ -599,7 +513,13 @@ static Bitu INT15_Handler(void) {
 		break;
 	case 0x83:	/* BIOS - SET EVENT WAIT INTERVAL */
 		{
-			if(reg_al == 0x01) LOG(LOG_BIOS,LOG_WARN)("Bios set event interval cancelled: not handled");   
+			if(reg_al == 0x01) { /* Cancel it */
+				mem_writeb(BIOS_WAIT_FLAG_ACTIVE,0);
+				IO_Write(0x70,0xb);
+				IO_Write(0x71,IO_Read(0x71)&~0x40);
+				CALLBACK_SCF(false);
+				break;
+			}
 			if (mem_readb(BIOS_WAIT_FLAG_ACTIVE)) {
 				reg_ah=0x80;
 				CALLBACK_SCF(true);
@@ -715,10 +635,6 @@ static Bitu INT15_Handler(void) {
 		CALLBACK_SCF(false);
 		reg_ah=0;
 		break;
-	case 0xc3:      /* set carry flag so BorlandRTM doesn't assume a VECTRA/PS2 */
-		reg_ah=0x86;
-		CALLBACK_SCF(true);
-		break;
 	case 0xc2:	/* BIOS PS2 Pointing Device Support */
 		switch (reg_al) {
 		case 0x00:		// enable/disable
@@ -777,6 +693,10 @@ static Bitu INT15_Handler(void) {
 			break;
 		}
 		break;
+	case 0xc3:      /* set carry flag so BorlandRTM doesn't assume a VECTRA/PS2 */
+		reg_ah=0x86;
+		CALLBACK_SCF(true);
+		break;
 	case 0xc4:	/* BIOS POS Programm option Select */
 		LOG(LOG_BIOS,LOG_NORMAL)("INT15:Function %X called, bios mouse not supported",reg_ah);
 		CALLBACK_SCF(true);
@@ -784,7 +704,11 @@ static Bitu INT15_Handler(void) {
 	default:
 		LOG(LOG_BIOS,LOG_ERROR)("INT15:Unknown call %4X",reg_ax);
 		reg_ah=0x86;
-		CALLBACK_SCF(false);
+		CALLBACK_SCF(true);
+		if ((machine==MCH_VGA) || (machine==MCH_CGA)) {
+			/* relict from comparisons, as int15 exits with a retf2 instead of an iret */
+			CALLBACK_SZF(false);
+		}
 	}
 	return CBRET_NONE;
 }
@@ -807,30 +731,29 @@ void BIOS_SetupDisks(void);
 
 class BIOS:public Module_base{
 private:
-	CALLBACK_HandlerObject callback[9];
+	CALLBACK_HandlerObject callback[10];
 public:
 	BIOS(Section* configuration):Module_base(configuration){
 		/* tandy DAC can be requested in tandy_sound.cpp by initializing this field */
 		bool use_tandyDAC=(real_readb(0x40,0xd4)==0xff);
+
 		/* Clear the Bios Data Area (0x400-0x5ff, 0x600- is accounted to DOS) */
 		for (Bit16u i=0;i<0x200;i++) real_writeb(0x40,i,0);
+
 		/* Setup all the interrupt handlers the bios controls */
 		/* INT 8 Clock IRQ Handler */
-		//TODO Maybe give this a special callback that will also call int 8 instead of starting 
-		//a new system
-		callback[0].Install(INT8_Handler,CB_IRET,"Int 8 Clock");
+		callback[0].Install(INT8_Handler,CB_IRQ0,"Int 8 Clock");
 		callback[0].Set_RealVec(0x8);
-		Bit16u call_int8=callback[0].Get_callback();
-		phys_writeb(CB_BASE+(call_int8<<4)+0,(Bit8u)0xFE);		//GRP 4
-		phys_writeb(CB_BASE+(call_int8<<4)+1,(Bit8u)0x38);		//Extra Callback instruction
-		phys_writew(CB_BASE+(call_int8<<4)+2,call_int8);		//The immediate word          
-		phys_writeb(CB_BASE+(call_int8<<4)+4,(Bit8u)0x50);		// push ax
-		phys_writeb(CB_BASE+(call_int8<<4)+5,(Bit8u)0xb0);		// mov al, 0x20
-		phys_writeb(CB_BASE+(call_int8<<4)+6,(Bit8u)0x20);
-		phys_writeb(CB_BASE+(call_int8<<4)+7,(Bit8u)0xe6);		// out 0x20, al
-		phys_writeb(CB_BASE+(call_int8<<4)+8,(Bit8u)0x20);
-		phys_writeb(CB_BASE+(call_int8<<4)+9,(Bit8u)0x58);		// pop ax
-		phys_writeb(CB_BASE+(call_int8<<4)+10,(Bit8u)0xcf);		// iret
+		// pseudocode for CB_IRQ0:
+		//	callback INT8_Handler
+		//	push ax,dx,ds
+		//	int 0x1c
+		//	cli
+		//	pop ds,dx
+		//	mov al, 0x20
+		//	out 0x20, al
+		//	pop ax
+		//	iret
 
 		mem_writed(BIOS_TIMER,0);			//Calculate the correct time
 
@@ -853,7 +776,7 @@ public:
 		BIOS_SetupDisks();
 
 		/* INT 14 Serial Ports */
-		callback[3].Install(&INT14_Handler,CB_IRET,"Int 14 COM-port");
+		callback[3].Install(&INT14_Handler,CB_IRET_STI,"Int 14 COM-port");
 		callback[3].Set_RealVec(0x14);
 
 		/* INT 15 Misc Calls */
@@ -872,13 +795,29 @@ public:
 		callback[6].Set_RealVec(0x1A);
 
 		/* INT 1C System Timer tick called from INT 8 */
-		callback[7].Install(&INT1C_Handler,CB_IRET,"Int 1c Timer tick");
+		callback[7].Install(&INT1C_Handler,CB_IRET,"Int 1c Timer");
 		callback[7].Set_RealVec(0x1C);
 		
 		/* IRQ 8 RTC Handler */
 		callback[8].Install(&INT70_Handler,CB_IRET,"Int 70 RTC");
 		callback[8].Set_RealVec(0x70);
 
+		/* Irq 9 rerouted to irq 2 */
+		callback[9].Install(NULL,CB_IRQ9,"irq 9 bios");
+		callback[9].Set_RealVec(0x71);
+
+		/* Irq 2 */
+		RealPt irq2pt=RealMake(0xf000,0xff55);	/* Ghost busters 2 mt32 mode */
+		Bitu call_irq2=CALLBACK_Allocate();	
+		CALLBACK_Setup(call_irq2,NULL,CB_IRET_EOI_PIC1,Real2Phys(irq2pt),"irq 2 bios");
+		RealSetVec(0x0a,irq2pt);
+
+		/* Some hardcoded vectors */
+		phys_writeb(0xfff53,0xcf);	/* bios default interrupt vector location */
+		phys_writeb(0xfe987,0xea);	/* original IRQ1 location (Defender booter) */
+		phys_writed(0xfe988,RealGetVec(0x09));
+		phys_writew(Real2Phys(RealGetVec(0x12))+0x12,0x20); //Hack for Jurresic
+
 		if (machine==MCH_TANDY) phys_writeb(0xffffe,0xff)	;	/* Tandy model */
 		else if (machine==MCH_PCJR) phys_writeb(0xffffe,0xfd);	/* PCJr model */
 		else phys_writeb(0xffffe,0xfc);	/* PC */
@@ -891,8 +830,20 @@ public:
 				real_writeb(0x40,0xd4,0xff);	/* tandy DAC init value */
 				real_writed(0x40,0xd6,0x00000000);
 				/* install the DAC callback handler */
-				tandy_DAC_callback=new CALLBACK_HandlerObject();
-				tandy_DAC_callback->Install(&IRQ_TandyDAC,CB_IRET,"Tandy DAC IRQ");
+				tandy_DAC_callback[0]=new CALLBACK_HandlerObject();
+				tandy_DAC_callback[1]=new CALLBACK_HandlerObject();
+				tandy_DAC_callback[0]->Install(&IRQ_TandyDAC,CB_IRET,"Tandy DAC IRQ");
+				tandy_DAC_callback[1]->Install(NULL,CB_TDE_IRET,"Tandy DAC end transfer");
+				// pseudocode for CB_TDE_IRET:
+				//	push ax
+				//	mov ax, 0x91fb
+				//	int 15
+				//	cli
+				//	mov al, 0x20
+				//	out 0x20, al
+				//	pop ax
+				//	iret
+
 				RealPt current_irq=RealGetVec(tandy_sb.irq+8);
 				real_writed(0x40,0xd6,current_irq);
 				for (Bitu i=0; i<0x10; i++) phys_writeb(PhysMake(0xf000,0xa084+i),0x80);
@@ -900,61 +851,63 @@ public:
 		}
 	
 		/* Setup some stuff in 0x40 bios segment */
+		
+		// port timeouts
+		// always 1 second even if the port does not exist
+		mem_writeb(BIOS_LPT1_TIMEOUT,1);
+		mem_writeb(BIOS_LPT2_TIMEOUT,1);
+		mem_writeb(BIOS_LPT3_TIMEOUT,1);
+		mem_writeb(BIOS_COM1_TIMEOUT,1);
+		mem_writeb(BIOS_COM2_TIMEOUT,1);
+		mem_writeb(BIOS_COM3_TIMEOUT,1);
+		mem_writeb(BIOS_COM4_TIMEOUT,1);
+		
 		/* detect parallel ports */
-	Bit8u DEFAULTPORTTIMEOUT = 10;	// 10 whatevers
-	Bitu ppindex=0; // number of lpt ports
-	if ((IO_Read(0x378)!=0xff)|(IO_Read(0x379)!=0xff)) {
-		// this is our LPT1
-		mem_writew(BIOS_ADDRESS_LPT1,0x378);
-		mem_writeb(BIOS_LPT1_TIMEOUT,DEFAULTPORTTIMEOUT);
-		ppindex++;
-		if((IO_Read(0x278)!=0xff)|(IO_Read(0x279)!=0xff)) {
-			// this is our LPT2
-			mem_writew(BIOS_ADDRESS_LPT2,0x278);
-			mem_writeb(BIOS_LPT2_TIMEOUT,DEFAULTPORTTIMEOUT);
+		Bitu ppindex=0; // number of lpt ports
+		if ((IO_Read(0x378)!=0xff)|(IO_Read(0x379)!=0xff)) {
+			// this is our LPT1
+			mem_writew(BIOS_ADDRESS_LPT1,0x378);
 			ppindex++;
-			if((IO_Read(0x3bc)!=0xff)|(IO_Read(0x3be)!=0xff)) {
-				// this is our LPT3
-				mem_writew(BIOS_ADDRESS_LPT3,0x3bc);
-				mem_writeb(BIOS_LPT3_TIMEOUT,DEFAULTPORTTIMEOUT);
+			if((IO_Read(0x278)!=0xff)|(IO_Read(0x279)!=0xff)) {
+				// this is our LPT2
+				mem_writew(BIOS_ADDRESS_LPT2,0x278);
+				ppindex++;
+				if((IO_Read(0x3bc)!=0xff)|(IO_Read(0x3be)!=0xff)) {
+					// this is our LPT3
+					mem_writew(BIOS_ADDRESS_LPT3,0x3bc);
+					ppindex++;
+				}
+			} else if((IO_Read(0x3bc)!=0xff)|(IO_Read(0x3be)!=0xff)) {
+				// this is our LPT2
+				mem_writew(BIOS_ADDRESS_LPT2,0x3bc);
 				ppindex++;
 			}
 		} else if((IO_Read(0x3bc)!=0xff)|(IO_Read(0x3be)!=0xff)) {
-			// this is our LPT2
-			mem_writew(BIOS_ADDRESS_LPT2,0x3bc);
-			mem_writeb(BIOS_LPT2_TIMEOUT,DEFAULTPORTTIMEOUT);
+			// this is our LPT1
+			mem_writew(BIOS_ADDRESS_LPT1,0x3bc);
+			ppindex++;
+			if((IO_Read(0x278)!=0xff)|(IO_Read(0x279)!=0xff)) {
+				// this is our LPT2
+				mem_writew(BIOS_ADDRESS_LPT2,0x278);
+				ppindex++;
+			}
+		} else if((IO_Read(0x278)!=0xff)|(IO_Read(0x279)!=0xff)) {
+			// this is our LPT1
+			mem_writew(BIOS_ADDRESS_LPT1,0x278);
 			ppindex++;
 		}
-	} else if((IO_Read(0x3bc)!=0xff)|(IO_Read(0x3be)!=0xff)) {
-		// this is our LPT1
-		mem_writew(BIOS_ADDRESS_LPT1,0x3bc);
-		mem_writeb(BIOS_LPT1_TIMEOUT,DEFAULTPORTTIMEOUT);
-		ppindex++;
-		if((IO_Read(0x278)!=0xff)|(IO_Read(0x279)!=0xff)) {
-			// this is our LPT2
-			mem_writew(BIOS_ADDRESS_LPT2,0x278);
-			mem_writeb(BIOS_LPT2_TIMEOUT,DEFAULTPORTTIMEOUT);
-			ppindex++;
-		}
-	}
-	else if((IO_Read(0x278)!=0xff)|(IO_Read(0x279)!=0xff)) {
-		// this is our LPT1
-		mem_writew(BIOS_ADDRESS_LPT1,0x278);
-		mem_writeb(BIOS_LPT1_TIMEOUT,DEFAULTPORTTIMEOUT);
-		ppindex++;
-	}
 
-	/* Setup equipment list */
-	// look http://www.bioscentral.com/misc/bda.htm
-	
-	//Bitu config=0x4400;	//1 Floppy, 2 serial and 1 parrallel 
-	Bitu config = 0x0;
-	
-	// set number of parallel ports
-	// if(ppindex == 0) config |= 0x8000; // looks like 0 ports are not specified
-	//else if(ppindex == 1) config |= 0x0000;
-	if(ppindex == 2) config |= 0x4000;
-	else config |= 0xc000;	// 3 ports
+		/* Setup equipment list */
+		// look http://www.bioscentral.com/misc/bda.htm
+		
+		//Bitu config=0x4400;	//1 Floppy, 2 serial and 1 parrallel 
+		Bitu config = 0x0;
+		
+		// set number of parallel ports
+		// if(ppindex == 0) config |= 0x8000; // looks like 0 ports are not specified
+		//else if(ppindex == 1) config |= 0x0000;
+		if(ppindex == 2) config |= 0x4000;
+		else config |= 0xc000;	// 3 ports
 #if (C_FPU)
 		//FPU
 		config|=0x2;
@@ -977,6 +930,8 @@ public:
 		}
 		// PS2 mouse
 		config |= 0x04;
+		// Gameport
+		config |= 0x1000;
 		mem_writew(BIOS_CONFIGURATION,config);
 		CMOS_SetRegister(0x14,config); //Should be updated on changes
 		/* Setup extended memory size */
@@ -985,10 +940,6 @@ public:
 		IO_Write(0x70,0x31);
 		size_extended|=(IO_Read(0x71) << 8);
 
-		phys_writeb(0xfff53,0xcf);	/* bios default interrupt vector location */
-		phys_writeb(0xfe987,0xea);	/* original IRQ1 location (Defender booter) */
-		phys_writed(0xfe988,RealGetVec(0x09));
-
 		if (machine==MCH_PCJR) PIC_AddEvent(RAMRefresh_Event,RAM_REFRESH_DELAY);
 	}
 	~BIOS(){
@@ -998,15 +949,17 @@ public:
 			IO_Write(tandy_sb.port+0xc,0xd0);
 		}
 		real_writeb(0x40,0xd4,0x00);
-		if (tandy_DAC_callback) {
+		if (tandy_DAC_callback[0]) {
 			Bit32u orig_vector=real_readd(0x40,0xd6);
-			if (orig_vector==tandy_DAC_callback->Get_RealPointer()) {
+			if (orig_vector==tandy_DAC_callback[0]->Get_RealPointer()) {
 				/* set IRQ vector to old value */
 				RealSetVec(tandy_sb.irq+8,real_readd(0x40,0xd6));
 				real_writed(0x40,0xd6,0x00000000);
 			}
-			delete tandy_DAC_callback;
-			tandy_DAC_callback=NULL;
+			delete tandy_DAC_callback[0];
+			delete tandy_DAC_callback[1];
+			tandy_DAC_callback[0]=NULL;
+			tandy_DAC_callback[1]=NULL;
 		}
 	}
 };
@@ -1014,24 +967,14 @@ public:
 // set com port data in bios data area
 // parameter: array of 4 com port base addresses, 0 = none
 void BIOS_SetComPorts(Bit16u baseaddr[]) {
-	Bit8u DEFAULTPORTTIMEOUT = 10;	// 10 whatevers
 	Bit16u portcount=0;
 	Bit16u equipmentword;
 	for(Bitu i = 0; i < 4; i++) {
 		if(baseaddr[i]!=0) portcount++;
-		if(i==0) {	// com1
-			mem_writew(BIOS_BASE_ADDRESS_COM1,baseaddr[i]);
-			mem_writeb(BIOS_COM1_TIMEOUT,DEFAULTPORTTIMEOUT);
-		} else if(i==1) {
-			mem_writew(BIOS_BASE_ADDRESS_COM2,baseaddr[i]);
-			mem_writeb(BIOS_COM2_TIMEOUT,DEFAULTPORTTIMEOUT);
-		} else if(i==2) {
-			mem_writew(BIOS_BASE_ADDRESS_COM3,baseaddr[i]);
-			mem_writeb(BIOS_COM3_TIMEOUT,DEFAULTPORTTIMEOUT);
-		} else {
-			mem_writew(BIOS_BASE_ADDRESS_COM4,baseaddr[i]);
-			mem_writeb(BIOS_COM4_TIMEOUT,DEFAULTPORTTIMEOUT);
-		}
+		if(i==0)		mem_writew(BIOS_BASE_ADDRESS_COM1,baseaddr[i]);
+		else if(i==1)	mem_writew(BIOS_BASE_ADDRESS_COM2,baseaddr[i]);
+		else if(i==2)	mem_writew(BIOS_BASE_ADDRESS_COM3,baseaddr[i]);
+		else			mem_writew(BIOS_BASE_ADDRESS_COM4,baseaddr[i]);
 	}
 	// set equipment word
 	equipmentword = mem_readw(BIOS_CONFIGURATION);
